//-----------------------------------------------------------------------------
// Copyright (c) 2005 Cypress Semiconductor, Inc. All rights reserved
//
// File: inand.c
//  Contents: 
//    Interleave NAND support the bInterLeave Flag will enable
//    Odd number NAND chips, the interleave FW will be disable
//    Even number NAND chips, then interleave FW will be enable
//
// Support: Both 512 NAND and 2K NAND
//    To support 2K NAND use the compiler switch NAND_2K, to support 4K NAND use both the compiler switches NAND_2K & NAND_4K
//
// indent 4.  NO TABS!
//--------------------------------------------------------------------------
//
// Copyright 2005, Cypress Semiconductor Corporation.
//
// This software is owned by Cypress Semiconductor Corporation (Cypress)
// and is protected by and subject to worldwide patent protection (United
// States and foreign), United States copyright laws and international 
// treaty provisions. Cypress hereby grants to licensee a personal, 
// non-exclusive, non-transferable license to copy, use, modify, create 
// derivative works of, and compile the Cypress Source Code and derivative 
// works for the sole purpose of creating custom software in support of   
// licensee product to be used only in conjunction with a Cypress integrated 
// circuit as specified in the applicable agreement. Any reproduction, 
// modification, translation, compilation, or representation of this 
// software except as specified above is prohibited without the express 
// written permission of Cypress.
//
// Disclaimer: CYPRESS MAKES NO WARRANTY OF ANY KIND,EXPRESS OR IMPLIED, 
// WITH REGARD TO THIS MATERIAL, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Cypress reserves the right to make changes without further notice to the
// materials described herein. Cypress does not assume any liability arising
// out of the application or use of any product or circuit described herein.
// Cypress does not authorize its products for use as critical components in
// life-support systems where a malfunction or failure may reasonably be
// expected to result in significant injury to the user. The inclusion of
// Cypress' product in a life-support systems application implies that the
// manufacturer assumes all risk of such use and in doing so indemnifies
// Cypress against all charges.
//
// Use may be limited by and subject to the applicable Cypress software
// license agreement.
//
//--------------------------------------------------------------------------
//
// Note:
//    This code has been optimized for Performance, so there are global variables
//    have been used instead of passing variables to subroutines.  
//      Any variables start with the prefix letter "g" indicates "Global" variable.
//
//    pragma OT(7, SPEED) should not be removed.  If remove, the compiler might generate
//    the code which will result slower performance
//
//    The design of this code does not follow Smart Media standard for the following:
//
//      1. Redundant structure
//               BYTE   data_status  = 0xff = good, else bad block
//               BYTE   boot_loader_status = 0xff = free block else used block
//               BYTE   ECC[6]
//               WORD   Address1, Address2
//               BYTE   Reserved[4] = 0xff
//
//      2. The "Address1" and "Address2" do not have parity bit.  The BuidTable does
//         support error bit checking via compare these address fields.  If not mach
//         if will continue to search for good address fields in the next page until
//         3 times of retry.  
//
//      3. Zone table size is 1000
//
// $Workfile: inand.c $
// $Date: 7/05/06 3:57p $
// $Revision: 3 $
// History:  May/12/05   Initial Release to customer
//-----------------------------------------------------------------------------
#include "globals.h"


xdata WORD gLog2Phy[cMaxBlock];         // bit0-11 addr map, bit12=1=free, 0=used
BYTE  gCurZone, gZones, gBankSel, gPartialCpy;  // Use this byte to reload the Table
WORD  gDst, gSrc, gFreeBlk;             // These variable will be shared

#ifdef NAND_2K
void WriteSpare()							
{												
        P_XGPIFSGLDATLX = MSB(gDst); _nop_();     // Phy2Log
        P_XGPIFSGLDATLX = LSB(gDst); _nop_();        
        P_XGPIFSGLDATLX = MSB(gDst); _nop_();        
        P_XGPIFSGLDATLX = LSB(gDst); _nop_();        
        P_XGPIFSGLDATLX = 0xFF;      _nop_();     // ST Micro fix
        P_XGPIFSGLDATLX = 0xFF;      _nop_();       
        P_XGPIFSGLDATLX = 0xFF;      _nop_();        
        P_XGPIFSGLDATLX = 0xFF;      _nop_();        
}
#endif

DWORD gSrcAdd, gPhyAdd, gNextLBA;       // sharing variables in multiple subrountines
BYTE  ecc0[6], ecc1[6];                 // Optimize ECC variables
BYTE  ecc2[6], ecc3[6];					// Added for 4K page 1024 sector size
BYTE  gSrcBlk0;                         // use for EraseBlock
xword *pDst;

static BYTE RetryCount = 0;
static BYTE DataFail; 

#ifndef USE_2NAND
// fast coding for bank select
const char code aBanks[4]={0xfc, 0xf3, 0xcf, 0x3f};  // interleave bank
const char code aBank0[4]={0xfe, 0xfb, 0xef, 0xbf};  // interleave bank
const char code aBank1[4]={0xfd, 0xf7, 0xdf, 0x7f};  // interleave bank
const char code nBank[8]={0xfe, 0xfd, 0xfb, 0xf7, 0xef, 0xdf, 0xbf, 0x7f};
BYTE  gEnableBanks;

#ifndef NAND_2K
BYTE  gEnableBank0, gEnableBank1;      // optimize variables
#endif

#endif


// Note: These variables below will be used to optimize speed
// These variables use in conjunction with MPAGE that map to 0xE6xx
pbyte P_GPIFTCB0        _at_ 0xd1;    // GPIFTCB0
pbyte P_GPIFTCB1        _at_ 0xd0;    // GPIFTCB1
pbyte P_XAUTODAT1       _at_ 0x7b;    // XAUTODAT1
pbyte P_XAUTODAT2       _at_ 0x7c;    // XAUTODAT2
pbyte P_ECC1B0[6]       _at_ 0x2a;    // 6 bytes ECC registers
pbyte P_INPKTEND        _at_ 0x48;    // INPKTEND
pbyte P_OUTPKTEND       _at_ 0x49;    // OUTPKTEND
pbyte P_ECCCFG          _at_ 0x28;    // ECCCFG
pbyte P_ECCRESET        _at_ 0x29;    // ECCRESET
pbyte P_EP6BCH          _at_ 0x98;    // EP6BCH
pbyte P_EP6BCL          _at_ 0x99;    // EP6BCL
pbyte P_EP4BCH          _at_ 0x94;    // EP4BCH
pbyte P_EP4BCL          _at_ 0x95;    // EP4BCL
pbyte P_GPIFREADYSTAT   _at_ 0xf4;    // GPIFREADYSTAT
pbyte P_FIFORESET       _at_ 0x04;    // FIFORESET
pbyte P_EP4CFG			_at_ 0x13;	  // EP4CFG
pbyte P_EP6CFG          _at_ 0x14;    // EP6CFG
pbyte P_EP2FIFOCFG      _at_ 0x18;    // EP2FIFOCFG
pbyte P_EP4FIFOCFG		_at_ 0x19;	  // EP4FIfOCFG
pbyte P_XGPIFSGLDATLX   _at_ 0xf1;    // XGPIFSGLDATLX
pbyte P_XGPIFSGLDATLNOX _at_ 0xf2;    // XGPIFSGLDATLNOX
pbyte P_EP2CS           _at_ 0xa3;    // EP2CS
pbyte P_EP4CS           _at_ 0xa4;    // EP4CS

#ifdef NAND_2K
bit CorrectData_EP4()
{    
    BYTE a,b,c,x,i;

    for (i=0;  i<2;  i++)
    {       
        b = ecc0[0] ^ ecc1[0];
        a = ecc0[1] ^ ecc1[1];
        c = ecc0[2] ^ ecc1[2];
        if (a|b|c)
        {
            if ((((a^(a>>1)) & 0x55) == 0x55) && // Does each pair of parity bits contain
                (((b^(b>>1)) & 0x55) == 0x55) && // one error and one match?
                 (((c^(c>>1)) & 0x54) == 0x54))
            {
                x = (a & 0x80); // If so, there's a one-bit error in data[].
                if (a & 0x20) x |= 0x40; // Find which byte is in error...
                if (a & 0x08) x |= 0x20;
                if (a & 0x02) x |= 0x10;
                if (b & 0x80) x |= 0x08;
                if (b & 0x20) x |= 0x04;
                if (b & 0x08) x |= 0x02;
                if (b & 0x02) x |= 0x01;
                a = 0; // ... and which bit...
                if (c & 0x80) a |= 0x04; 
                if (c & 0x20) a |= 0x02;
                if (c & 0x08) a |= 0x01;
                EP4FIFOBUF[x+ (WORD)(i<<8)] ^= (1 << a); // ... and correct it.
                // one bit is ok
            }
            else { b2BitErr=1; return cFail; }  //  two bits return error
        }
        ecc0[0] = ecc0[3];
        ecc0[1] = ecc0[4];
        ecc0[2] = ecc0[5];
        ecc1[0] = ecc1[3];
        ecc1[1] = ecc1[4];
        ecc1[2] = ecc1[5];
    }
    return cOK;
}
#endif
//==========================================================================
// NAND Hardware/software init
//   Donot call this subroutine during the SOFT RESET (i.e. softReset)
//==========================================================================
void InitNAND()
{
#ifdef NAND_2K
	BYTE tmpBank;
#endif
	gNandBits = 0;      // all the NAND bits variables = 0
    FwCfg = 0;
    NandCfg = 0;
    gFreeBlk = 0; 
    gPartialCpy = 0;
    dwLBA = 0;
#ifdef NAND_2K
	for(tmpBank=0;tmpBank<8;tmpBank++)											// Issue RESET for all banks
	{
		nand_ready3();
		DISABLE_NAND();
		IOD = nBank[tmpBank];
		NandSendCmd(cNAND_RESET); _nop_();_nop_();		// Reset fix for Micron 4k page NANDS
	}				
	nand_ready3();
#endif
    GetNandCfg();       // Get NAND Configuration from MFG TOOL
    directionIn=1;      // Compute Log2Phy Table at init time
    NAND_WP = 1;
    gCurZone = 0xff;    // Load current zone
    Log2Phy();
    ready_default();    // restore default value
#ifndef NO_WP
    bWPSwitchState = NAND_WP_SWITCH;
#endif
#ifdef NAND_2K
	nand_ready3();
#endif
    DISABLE_NAND();    
}

//==========================================================================
// Get Free block from LUT
// This search always return a free block
//==========================================================================
void nGetFreeBlk()
{
    while (!bFreeFound)          // search until found free block
        nSearchFreeBlock(0);     // will search for 256 entries
    bFreeFound = 0;
}

//==========================================================================
// NAND EraseBlocks on both banks
//  input: gSrcAdd     = Current Src Block
//==========================================================================
void nEraseBlock()
{
    if (bNeedErase)              // only erase if needed
    {   // erase both banks in the interleave mode

#ifdef USE_2NAND
		if (bInterLeave)
        	CE0_ON(), CE1_ON();      // Enable both banks for interleave mode
#else
        IOD=gEnableBanks;        // non-interleave this should be only 1 bank
#endif
        xSrcAdd = gSrcBlk0;
#ifdef NAND_2K
		NandSendCmd(cNAND_RESET); _nop_();_nop_();				
#endif
        nand_blk_erase(gSrcAdd);    
        // after erasing the NAND, there are some spare time for background search
        nSearchFreeBlock(0);     // will search for 256 entries
    }
}

#pragma OT(7, SPEED)            // Use this optimize for the speed

//==========================================================================
// Clear 16-byte each pass.  
// Maximun 4K buffer 
//==========================================================================
void memset16(BYTE xdata *dest, BYTE c, BYTE len)
{
    AUTOPTR1H = MSB(dest); AUTOPTR1L = LSB(dest);
    do
    {    // Unrole the loop to speed up memory fill
         P_XAUTODAT1 = P_XAUTODAT1 = P_XAUTODAT1 = P_XAUTODAT1 = 
         P_XAUTODAT1 = P_XAUTODAT1 = P_XAUTODAT1 = P_XAUTODAT1 = 
         P_XAUTODAT1 = P_XAUTODAT1 = P_XAUTODAT1 = P_XAUTODAT1 = 
         P_XAUTODAT1 = P_XAUTODAT1 = P_XAUTODAT1 = P_XAUTODAT1 = c;
    } while (--len>0);
}


//==========================================================================
// Setup AUTO pointer for P_XAUTODAT1
// Read ECC registers when first 256 ECC calculation done
// Read 512 byte data and save:
//     ecc0[] contains GPIF HW generate 6 ECC bytes
//     ecc1[] contains ECC data from redundant area
//==========================================================================
#ifdef NAND_4K

void ECCSetup_EP4()
{

	BYTE dummy_read;
	RetryCount = 5;
	DataFail = 1;

	dummy_read = P_XGPIFSGLDATLX;	// dummy offset
	dummy_read = P_XGPIFSGLDATLX;	// first byte of spare area of first half
	dummy_read = P_XGPIFSGLDATLX;
	ecc1[0]    = P_XGPIFSGLDATLX;
	ecc1[1]    = P_XGPIFSGLDATLX;
	ecc1[2]    = P_XGPIFSGLDATLX;
	ecc1[3]    = P_XGPIFSGLDATLX;
	ecc1[4]    = P_XGPIFSGLDATLX;
	ecc1[5]    = P_XGPIFSGLDATLX;
	dummy_read = P_XGPIFSGLDATLX;
	dummy_read = P_XGPIFSGLDATLX;
	dummy_read = P_XGPIFSGLDATLX;
	dummy_read = P_XGPIFSGLDATLX;
	dummy_read = P_XGPIFSGLDATLX;
	dummy_read = P_XGPIFSGLDATLX;
	dummy_read = P_XGPIFSGLDATLX;
	dummy_read = P_XGPIFSGLDATLX;
	dummy_read = P_XGPIFSGLDATLX;	// first byte of spare area of second half
	dummy_read = P_XGPIFSGLDATLX;
	ecc3[0]    = P_XGPIFSGLDATLX;
	ecc3[1]    = P_XGPIFSGLDATLX;
	ecc3[2]    = P_XGPIFSGLDATLX;
	ecc3[3]    = P_XGPIFSGLDATLX;
	ecc3[4]    = P_XGPIFSGLDATLX;
	ecc3[5]    = P_XGPIFSGLDATLX;

	if ( ((ecc0[0] != ecc1[0]) || (ecc0[1] != ecc1[1]) || (ecc0[2] != ecc1[2]) ||
    	  (ecc0[3] != ecc1[3]) || (ecc0[4] != ecc1[4]) || (ecc0[5] != ecc1[5]) ||
		  (ecc2[0] != ecc3[0]) || (ecc2[1] != ecc3[1]) || (ecc2[2] != ecc3[2]) ||
    	  (ecc2[3] != ecc3[3]) || (ecc2[4] != ecc3[4]) || (ecc2[5] != ecc3[5])) )
    {	 									// condition TRUE implies one or multiple bit error

  		P_FIFORESET = 0x04; _nop_();
		P_FIFORESET = 0x00; _nop_();

		do{
			if (CorrectData_EP4())				// condition TRUE implies incorrectible error, retry reading till Max retry counts
			{
				RetryCount--;
				NandSetAdd(cNAND_READ_DATA, xLBA3&SectorbCntMsk);

				P_GPIFTCB1 = MSB(511);
				P_ECCRESET = P_GPIFTCB0 = LSB(511);
				while(!(P_EP4CS&bmEP4EMPTY));	_nop_();
				P_FIFORESET = 0x04;			_nop_();
				P_EP4CFG = EP4CFG_DEFAULT;
				P_EP4FIFOCFG = 0x00;			_nop_();
				P_FIFORESET = 0x04;			_nop_();

				GPIFTRIG = 0x04 | cEP4;

				_nop_();
				while (P_GPIFTCB1);          // Read the first 256 ECC registers
	    		ecc0[0] = P_ECC1B0[0];       // save to local buffers
	    		ecc0[1] = P_ECC1B0[1]; 
	    		ecc0[2] = P_ECC1B0[2]; 
	    		while(!(gpifIdle()));		 // All 512 bytes have been read
	    		ecc0[3] = P_ECC1B0[3];       // save ECC 
	    		ecc0[4] = P_ECC1B0[4];       // GPIF will be done afer reading these
	    		ecc0[5] = P_ECC1B0[5];       // registers

				NandSetAdd(cNAND_READ_DATA, xLBA3&SectorbCntMsk + SpareOffset);
				dummy_read = P_XGPIFSGLDATLX;
				dummy_read = P_XGPIFSGLDATLX;
				dummy_read = P_XGPIFSGLDATLX;
				ecc1[0]    = P_XGPIFSGLDATLX;
				ecc1[1]    = P_XGPIFSGLDATLX;
				ecc1[2]    = P_XGPIFSGLDATLX;
				ecc1[3]    = P_XGPIFSGLDATLX;
				ecc1[4]    = P_XGPIFSGLDATLX;
				ecc1[5]    = P_XGPIFSGLDATLX;

				if( (ecc0[0] != ecc1[0]) || (ecc0[1] != ecc1[1]) || (ecc0[2] != ecc1[2]) ||
         			(ecc0[3] != ecc1[3]) || (ecc0[4] != ecc1[4]) || (ecc0[5] != ecc1[5]) )
     			{	 
					continue;
				} else DataFail = 0;
			}
			else DataFail = 0;					// one bit error fixed by CorrectData_EP4
		} while (RetryCount && DataFail);
		if (DataFail)
		{ bErr = 1; bCnt = 0; goto EP4ECCend;}
		if(EZUSB_HIGHSPEED())
		{
			P_EP4BCH = 0x02;
			P_EP4BCL = 0x00;
		}
		else
		{
			while(!gpifIdle());
			NandSetAdd(cNAND_READ_DATA, xLBA3&SectorbCntMsk);
			while(!(P_EP4CS&bmEP4EMPTY)); _nop_();
			P_FIFORESET = 0x04; _nop_();
			P_EP4CFG = EP4CFG_DEFAULT;
			P_EP4FIFOCFG = (bmINFM | bmAUTOIN);

			EP4AUTOINLENH = MSB(64);
			EP4AUTOINLENL = LSB(64);

			P_GPIFTCB1 = MSB(527);
			P_GPIFTCB0 = LSB(527);

			GPIFTRIG = 0x04 | cEP4; _nop_();
			while(!gpifIdle());
			while(!(P_EP4CS&bmEP4EMPTY));
			P_INPKTEND = 0x84;
			P_EP4FIFOCFG = 0;
		}

		RetryCount = 5; DataFail = 1;
		do{
			while(!gpifIdle());
			NandSetAdd(cNAND_READ_DATA, (xLBA3&SectorbCntMsk)+SecondHalf);

			P_GPIFTCB1 = MSB(511);
			P_ECCRESET = P_GPIFTCB0 = LSB(511);
			while(!(P_EP4CS&bmEP4EMPTY));	_nop_();
			P_FIFORESET = 0x04;			_nop_();
			P_EP4CFG = EP4CFG_DEFAULT;
			P_EP4FIFOCFG = 0x00;			_nop_();
			P_FIFORESET = 0x04;			_nop_();
			P_INPKTEND = 0x84;          _nop_();
			P_INPKTEND = 0x84;          _nop_();

			GPIFTRIG = 0x04 | cEP4;

			_nop_();
			while (P_GPIFTCB1);          // Read the first 256 ECC registers
    		ecc0[0] = P_ECC1B0[0];       // save to local buffers
    		ecc0[1] = P_ECC1B0[1]; 
    		ecc0[2] = P_ECC1B0[2]; 
    		while(!(gpifIdle()));		 // All 512 bytes have been read
    		ecc0[3] = P_ECC1B0[3];       // save ECC 
    		ecc0[4] = P_ECC1B0[4];       // GPIF will be done afer reading these
    		ecc0[5] = P_ECC1B0[5];       // registers

			NandSetAdd(cNAND_READ_DATA, (xLBA3&SectorbCntMsk) + SpareOffset+SecondHalf);
			dummy_read = P_XGPIFSGLDATLX;
			dummy_read = P_XGPIFSGLDATLX;
			dummy_read = P_XGPIFSGLDATLX;
			ecc1[0]    = P_XGPIFSGLDATLX;
			ecc1[1]    = P_XGPIFSGLDATLX;
			ecc1[2]    = P_XGPIFSGLDATLX;
			ecc1[3]    = P_XGPIFSGLDATLX;
			ecc1[4]    = P_XGPIFSGLDATLX;
			ecc1[5]    = P_XGPIFSGLDATLX;

			if( (ecc0[0] != ecc1[0]) || (ecc0[1] != ecc1[1]) || (ecc0[2] != ecc1[2]) ||
         		(ecc0[3] != ecc1[3]) || (ecc0[4] != ecc1[4]) || (ecc0[5] != ecc1[5]) )
     			{	 
	 				if(CorrectData_EP4())
					{
						RetryCount--;
					}else DataFail = 0;
				} else DataFail = 0;

		} while (RetryCount && DataFail);
		if (DataFail)
		{ bErr = 1; bCnt = 0; goto EP4ECCend;}
		if(EZUSB_HIGHSPEED())
		{
			P_EP4BCH = 0x02;
			P_EP4BCL = 0x00;
		}
		else
		{
			while(!gpifIdle());
			NandSetAdd(cNAND_READ_DATA, xLBA3&SectorbCntMsk);
			while(!(P_EP4CS&bmEP4EMPTY)); _nop_();
			P_FIFORESET = 0x04; _nop_();
			P_EP4CFG = EP4CFG_DEFAULT;
			P_EP4FIFOCFG = (bmINFM | bmAUTOIN);

			EP4AUTOINLENH = MSB(64);
			EP4AUTOINLENL = LSB(64);

			P_GPIFTCB1 = MSB(527);
			P_GPIFTCB0 = LSB(527);

			GPIFTRIG = 0x04 | cEP4; _nop_();
			while(!gpifIdle());
			while(!(P_EP4CS&bmEP4EMPTY));
			P_INPKTEND = 0x84;
			P_EP4FIFOCFG = 0;
		}
		goto EP4ECCend;
	}
	else						// there was no error found in first read try, commit 1024 bytes
	{
		if(EZUSB_HIGHSPEED())
		{
			P_FIFORESET = 0x04; _nop_();
			P_FIFORESET = 0x00; _nop_();
			P_EP4BCH = 0x02;
			P_EP4BCL = 0x00;
			P_EP4BCH = 0x02;
			P_EP4BCL = 0x00;
		}
		else				// After one successful read, there is no need to correct data, hence reload the buffers for full speed after changing configurations
	    {		
				while(!(gpifIdle()));

				NandSetAdd(cNAND_READ_DATA, xLBA3&SectorbCntMsk);
				while(!(P_EP4CS&bmEP4EMPTY));
				_nop_();
				P_FIFORESET = 0x04;
				_nop_();
				P_EP4CFG =  EP4CFG_DEFAULT;
				P_EP4FIFOCFG = (bmINFM | bmAUTOIN);
				P_FIFORESET = 0x00; _nop_();
			
	   			EP4AUTOINLENH = MSB(64);
				EP4AUTOINLENL = LSB(64);

				P_GPIFTCB1 = MSB(1056);              // Setup GPIF count high
	            P_GPIFTCB0 = LSB(1056);

				GPIFTRIG = 0x04 | cEP4; _nop_();
				while(!(gpifIdle()));
				while(!(P_EP4CS&bmEP4EMPTY));
				P_INPKTEND = 0x84;
				P_EP4FIFOCFG = 0;
		}
	}

	EP4ECCend:
	dwLBA++;                
   	bCnt = --gSectorcount && !bErr;
   	bReload   = xLBA3 ==0 && bCnt;    // compute bReload
   	bMsk      = xLBA3 & SectorbCntMsk;   
   	if (!bMsk) xPhyAdd++;             // 2K NAND address inc
	else NandSetAdd(cNAND_READ_DATA, xLBA3&SectorbCntMsk);		// helps eliminate the unnecessary 8 dummy value reads
}

#else
#ifdef NAND_2K
void ECCSetup_EP4()
{
	BYTE dummy_read;

    // Loading the Pointer of stored 6-bytes ECC data

	while (P_GPIFTCB1);          // Read the first 256 ECC registers
    ecc0[0] = P_ECC1B0[0];       // save to local buffers
    ecc0[1] = P_ECC1B0[1]; 
    ecc0[2] = P_ECC1B0[2]; 
    while(!(gpifIdle()));		 // All 512 bytes have been read
//    dummy_read = P_XGPIFSGLDATLX;
    ecc0[3] = P_ECC1B0[3];       // save ECC 
    ecc0[4] = P_ECC1B0[4];       // GPIF will be done afer reading these
    ecc0[5] = P_ECC1B0[5];       // registers

	
    // Stored 6-byte ECC data from NAND Flash
	dummy_read = P_XGPIFSGLDATLX;		// 512th byte
	dummy_read = P_XGPIFSGLDATLNOX;		// 512th byte
	dummy_read = P_XGPIFSGLDATLX;		// 513th byte
	dummy_read = P_XGPIFSGLDATLNOX;		// 513th byte
	ecc1[0] = P_XGPIFSGLDATLX;    		// 514th byte
	ecc1[0] = P_XGPIFSGLDATLNOX;    	// 514th byte
    ecc1[1] = P_XGPIFSGLDATLX;    
    ecc1[1] = P_XGPIFSGLDATLNOX;    
    ecc1[2] = P_XGPIFSGLDATLX;    
    ecc1[2] = P_XGPIFSGLDATLNOX;    
    ecc1[3] = P_XGPIFSGLDATLX;    
    ecc1[3] = P_XGPIFSGLDATLNOX;    
    ecc1[4] = P_XGPIFSGLDATLX;    
    ecc1[4] = P_XGPIFSGLDATLNOX;    
    ecc1[5] = P_XGPIFSGLDATLX;
    ecc1[5] = P_XGPIFSGLDATLNOX;
	
	
    dummy_read = P_XGPIFSGLDATLX;
    dummy_read = P_XGPIFSGLDATLNOX;
    dummy_read = P_XGPIFSGLDATLX;
    dummy_read = P_XGPIFSGLDATLNOX;
    dummy_read = P_XGPIFSGLDATLX;
    dummy_read = P_XGPIFSGLDATLNOX;
    dummy_read = P_XGPIFSGLDATLX;
    dummy_read = P_XGPIFSGLDATLNOX;
    dummy_read = P_XGPIFSGLDATLX;
	dummy_read = P_XGPIFSGLDATLNOX;
	dummy_read = P_XGPIFSGLDATLX;
	dummy_read = P_XGPIFSGLDATLNOX;
	dummy_read = P_XGPIFSGLDATLX;
	dummy_read = P_XGPIFSGLDATLNOX;
	dummy_read = P_XGPIFSGLDATLX;
	dummy_read = P_XGPIFSGLDATLNOX;
	

	 if( ((ecc0[0] != ecc1[0]) || (ecc0[1] != ecc1[1]) || (ecc0[2] != ecc1[2]) ||
         (ecc0[3] != ecc1[3]) || (ecc0[4] != ecc1[4]) || (ecc0[5] != ecc1[5])) )
     {	 
	 	if(CorrectData_EP4())
		{	RetryCount = 5;
			DataFail = 1;

			do
			{
					
			NandSetAdd(cNAND_READ_DATA, xLBA3&SectorbCntMsk);
			P_GPIFTCB1 = MSB(511);              // Setup GPIF count high
            P_ECCRESET = P_GPIFTCB0 = LSB(511);
			while(!(P_EP4CS&bmEP4EMPTY));
			_nop_();
			P_FIFORESET = 0x04;
			_nop_();
			P_EP4CFG =  EP4CFG_DEFAULT;
			P_EP4FIFOCFG = 0x00;

			_nop_();
			P_FIFORESET = 0x04;
			_nop_();
			GPIFTRIG = 0x04 | cEP4;

			while (P_GPIFTCB1);          // Read the first 256 ECC registers
		    ecc0[0] = P_ECC1B0[0];       // save to local buffers
		    ecc0[1] = P_ECC1B0[1]; 
		    ecc0[2] = P_ECC1B0[2]; 
		    while(!(gpifIdle()));
//		    dummy_read = P_XGPIFSGLDATLX;
		    ecc0[3] = P_ECC1B0[3];       // save ECC 
		    ecc0[4] = P_ECC1B0[4];       // GPIF will be done afer reading these
		    ecc0[5] = P_ECC1B0[5];       // registers
		
			
		    // Stored 6-byte ECC data from NAND Flash
		    dummy_read = P_XGPIFSGLDATLX;
			dummy_read = P_XGPIFSGLDATLNOX;
			dummy_read = P_XGPIFSGLDATLX;
			dummy_read = P_XGPIFSGLDATLNOX;

			ecc1[0] = P_XGPIFSGLDATLX;    
			ecc1[0] = P_XGPIFSGLDATLNOX;    
		    ecc1[1] = P_XGPIFSGLDATLX;    
		    ecc1[1] = P_XGPIFSGLDATLNOX;    
		    ecc1[2] = P_XGPIFSGLDATLX;    
		    ecc1[2] = P_XGPIFSGLDATLNOX;    
		    ecc1[3] = P_XGPIFSGLDATLX;    
		    ecc1[3] = P_XGPIFSGLDATLNOX;    
		    ecc1[4] = P_XGPIFSGLDATLX;    
		    ecc1[4] = P_XGPIFSGLDATLNOX;    
		    ecc1[5] = P_XGPIFSGLDATLX;
		    ecc1[5] = P_XGPIFSGLDATLNOX;
			
			
			dummy_read = P_XGPIFSGLDATLX;
			dummy_read = P_XGPIFSGLDATLNOX;
			dummy_read = P_XGPIFSGLDATLX;
			dummy_read = P_XGPIFSGLDATLNOX;
			dummy_read = P_XGPIFSGLDATLX;
			dummy_read = P_XGPIFSGLDATLNOX;
			dummy_read = P_XGPIFSGLDATLX;
			dummy_read = P_XGPIFSGLDATLNOX;
			dummy_read = P_XGPIFSGLDATLX;
			dummy_read = P_XGPIFSGLDATLNOX;
			dummy_read = P_XGPIFSGLDATLX;
			dummy_read = P_XGPIFSGLDATLNOX;
			dummy_read = P_XGPIFSGLDATLX;
			dummy_read = P_XGPIFSGLDATLNOX;
			dummy_read = P_XGPIFSGLDATLX;            
			dummy_read = P_XGPIFSGLDATLNOX;
			

			if( (ecc0[0] != ecc1[0]) || (ecc0[1] != ecc1[1]) || (ecc0[2] != ecc1[2]) ||
         		(ecc0[3] != ecc1[3]) || (ecc0[4] != ecc1[4]) || (ecc0[5] != ecc1[5]) )
     			{	 
	 				if(CorrectData_EP4())
					{
							RetryCount--;
					}else DataFail = 0;
				} else DataFail = 0;

			} while(RetryCount&&DataFail);
					
			if(DataFail)
			{
				bErr = 1; bCnt = 0;	
			}
		}
     }
	 if(EZUSB_HIGHSPEED())
	 {
	 	P_EP4BCH = 0x02;
	 	P_EP4BCL = 0x00;
	 }
	else
    {		
			while(!(gpifIdle()));
					
			NandSetAdd(cNAND_READ_DATA, xLBA3&SectorbCntMsk);
			while(!(P_EP4CS&bmEP4EMPTY));
			_nop_();
			P_FIFORESET = 0x04;
			_nop_();
			P_EP4CFG =  EP4CFG_DEFAULT;
			P_EP4FIFOCFG = (bmINFM | bmAUTOIN);
			
   			EP4AUTOINLENH = MSB(64);
			EP4AUTOINLENL = LSB(64);

			P_GPIFTCB1 = MSB(527);              // Setup GPIF count high
            P_GPIFTCB0 = LSB(527);

			GPIFTRIG = 0x04 | cEP4; _nop_();
			while(!(gpifIdle()));
			while(!(P_EP4CS&bmEP4EMPTY));
			P_INPKTEND = 0x84;
			P_EP4FIFOCFG = 0;

	}
	dwLBA++;                
    bCnt = --gSectorcount && !bErr;
    bReload   = xLBA3 ==0 && bCnt;    // compute bReload
    bMsk      = xLBA3 & SectorbCntMsk;   
    if (!bMsk) xPhyAdd++;             // 2K NAND address inc
}
#endif
#endif
//==========================================================================
// Read NAND Pages: Supports for both interleave and non-interleave
//   return bErr and b2BitErr, which will send STALL to USB Host
// Modify:
//    ecc0[] and ecc1[]
//    bMsk, bCnt
//==========================================================================
void nReadPages()
{
#ifdef NAND_4K

	while(!(P_EP4CS&bmEP4EMPTY));					// wait for host to read all the data, this gives clean endpoints

	Log2Phy();

	do
	{
		while(!(P_EP4CS&bmEP4EMPTY));

		_nop_();
		P_FIFORESET = 0x04;        _nop_();			// EP4 changed to manual mode to set configuration
		P_EP4CFG = EP4CFG_DEFAULT;					// Also we need to save the first half (512 bytes of 1024 bytes) without reading out to host
		P_EP4FIFOCFG = 0x00;		 _nop_();
		P_FIFORESET = 0x04;        _nop_();
		P_FIFORESET = 0x80;      _nop_();


		P_GPIFTCB1 = MSB(511);						// Read first half of the sector
		P_ECCRESET = P_GPIFTCB0 = LSB(511);			// For error correction purposes, it is required to have transactions with max size 512 bytes - hardware limitation
		GPIFTRIG = 0x04 | cEP4;

		_nop_();
		while (P_GPIFTCB1);          // Read the first 256 ECC registers
    	ecc0[0] = P_ECC1B0[0];       // save to local buffers
    	ecc0[1] = P_ECC1B0[1]; 
    	ecc0[2] = P_ECC1B0[2]; 
    	while(!(gpifIdle()));		 // All 512 bytes have been read
    	ecc0[3] = P_ECC1B0[3];       // save ECC 
    	ecc0[4] = P_ECC1B0[4];       // GPIF will be done afer reading these
    	ecc0[5] = P_ECC1B0[5];       // registers

		P_INPKTEND = 0x04; _nop_();				// Commit the packet that has been read

												

		P_GPIFTCB1 = MSB(511);					// Read second half of the sector
		P_ECCRESET = P_GPIFTCB0 = LSB(511);
		GPIFTRIG = 0x04 | cEP4;

		_nop_();
		while (P_GPIFTCB1);          // Read the first 256 ECC registers
    	ecc2[0] = P_ECC1B0[0];       // save to local buffers
    	ecc2[1] = P_ECC1B0[1]; 
    	ecc2[2] = P_ECC1B0[2]; 
    	while(!(gpifIdle()));		 // All 512 bytes have been read
    	ecc2[3] = P_ECC1B0[3];       // save ECC 
    	ecc2[4] = P_ECC1B0[4];       // GPIF will be done afer reading these
    	ecc2[5] = P_ECC1B0[5];       // registers

		P_INPKTEND = 0x04; _nop_();				// Commit the second packet that has been read, this makes the sector ready for error check

		ECCSetup_EP4();

		if (!bMsk)
		{
			if (bReload)
			{
				Log2Phy();
			}
			else
				n2k_set_radd();
		}
	} while (bCnt);

#else
#ifdef NAND_2K
                                          // remap logical blocks
    // Read the first 512 byte data and save all the ECC data in ecc0[]
    // and ecc1[] buffers.  The CheckECC will be called to verify the ECC
    // during the polling for GPIF DMA done (USB data transfer to USB Host)
   	while(!(P_EP4CS&bmEP4EMPTY)); //wait for host to read all the data. 
	
	Log2Phy();

	_nop_();
	P_FIFORESET = 0x04;					//changing EP4 to manual mode
	_nop_();
	P_EP4CFG =  EP4CFG_DEFAULT;
	P_EP4FIFOCFG = 0x00;
	_nop_();
	P_FIFORESET = 0x04;
	_nop_();
	
	
    P_GPIFTCB1 = MSB(511);                  // Setup GPIF count high, read 1 sector
    P_ECCRESET = P_GPIFTCB0 = LSB(511);     // Reset ECC and GPIF low count
    
	GPIFTRIG = 0x04 | cEP4;                         //Trigger GPIF to EP4
    
	ECCSetup_EP4();
    
	if (!bMsk)
    {
        if (bReload)  // check if need to remap page
		{
			Log2Phy();
		}
        else n2k_set_radd();
    }
    if (bCnt)
    {
        do
        {	
            P_GPIFTCB1 = MSB(511);              // Setup GPIF count high
            P_ECCRESET = P_GPIFTCB0 = LSB(511); // Reset ECC and GPIF low count

			while(!(P_EP4CS&bmEP4EMPTY));
			_nop_();
			P_FIFORESET = 0x04;
			_nop_();
			P_EP4CFG =  EP4CFG_DEFAULT;
			P_EP4FIFOCFG = 0x00;

			_nop_();
			P_FIFORESET = 0x04;
			_nop_();


			
            GPIFTRIG = 0x04 | cEP4;                     // Arm EP4
            //=========================================================================
            // local computation while polling GPIF done to optimize the speed
            //=========================================================================
            
            ECCSetup_EP4();
            if (!bMsk) 
            {
                if(bReload)
				{
					Log2Phy(); 
				}
                else n2k_set_radd();
            }
        } while (bCnt);
    }
	
    
#else		//------------------------------512 BYTE code------------------------------
    Log2Phy();                               // remap logical blocks
    if (bInterLeave)                         // interleave FW
    {
        bank_select(cBank0);                 // select bank 0
        if (bLBA0)                           // if Odd sector, need aligment
        {
            FifoRd(cEP6, 528);               // dummy read-> next sector
            bank_select(cBank1);
        }
    }
    // Read the first 512 byte data and save all the ECC data in ecc0[]
    // and ecc1[] buffers.  The CheckECC will be called to verify the ECC
    // during the polling for GPIF DMA done (USB data transfer to USB Host)
    P_GPIFTCB1 = MSB(cNAND_PSIZE);               // Setup GPIF count high
    P_ECCRESET = P_GPIFTCB0 = LSB(cNAND_PSIZE);  // Reset ECC and GPIF low count
    GPIFTRIG = 0x04 | cEP4;                      // Arm EP4
    // Compute while GPIF Busy
    ECCSetup(EP4FIFO);
    if (bReload) { CheckECC(); Log2Phy(); }      // check if need to remap page
    if (bCnt)
    {
       AUTOPTRH2 = MSB(cGPIFBANK);
       AUTOPTRL2 = LSB(cGPIFBANK);
       do
       {
            P_GPIFTCB1 = MSB(cNAND_PSIZE);               // Setup GPIF count high
            P_ECCRESET = P_GPIFTCB0 = LSB(cNAND_PSIZE);  // Reset ECC and GPIF low count
            fast_bank_sel(gBankSel);                     // use AUTO PTR2
            GPIFTRIG = 0x04 | cEP4;                      // Arm EP4
            AUTOPTRL2 = LSB(cGPIFBANK);                  // init again
            //=========================================================================
            // local computation while polling GPIF done to optimize the speed
            //=========================================================================
            CheckECC();
            ECCSetup(EP4FIFO);
            if (bReload) Log2Phy(); 
        }  while (bCnt);
        AUTOPTRH2 = MSB(EP4FIFOBUF);
    }
    CheckECC();                              // if ECC error, will send stall
    bank_default();
#endif		// ifdef nand 2k

#endif		// ifdef nand 4k
#ifdef NAND_2K
	P_EP4FIFOCFG = EP4FIFOCFG_DEFAULT; 
#endif
    DISABLE_NAND();
}

//==========================================================================
// Setup AUTO pointer for P_XAUTODAT1
// Read ECC registers when first 256 ECC calculation done
// Read 512 byte data and save:
//     ecc0[] contains GPIF HW generate 6 ECC bytes
//     ecc1[] contains ECC data from redundant area
//==========================================================================
void ECCSetup(WORD offset)
{
    dwLBA++;                // next sector     
    bCnt = --gSectorcount && !bErr;
    // Loading the Pointer of stored 6-bytes ECC data
    AUTOPTR1H = MSB(offset);
    AUTOPTR1L = LSB(offset);
#ifdef NAND_2K
    bReload   = xLBA3 ==0 && bCnt;    // compute bReload
    bMsk      = xLBA3 & SectorbCntMsk;   
	if (!bMsk) xPhyAdd++;             // 2K NAND address inc
#else
    if (bInterLeave)
    {   
        // Set this flag when all the pages in the block are done
        // Interleave total pages are 64
        // Non-Interleave total pages are 32
        bReload   = (xLBA3 & cInterLeaveMsk)==0 && bCnt;    // compute bReload
        // Select bank base on the dwLBA bit 0
        if (bLBA0) gBankSel = cBank1; else gBankSel = cBank0;
    }
    else
    {
        gBankSel = cBank3;     // non interleave keep the default value
        bReload   = (xLBA3 & (c512PageSize-1))==0 && bCnt;    // compute bReload
    }
#endif
    while (P_GPIFTCB1);          // Read the first 256 ECC registers
    ecc0[0] = P_ECC1B0[0];       // save to local buffers
    ecc0[1] = P_ECC1B0[1]; 
    ecc0[2] = P_ECC1B0[2]; 
    while (P_GPIFTCB0 & 0xf8);   // Read second 256 ECC registers       
    ecc0[3] = P_ECC1B0[3];       // save ECC 
    ecc0[4] = P_ECC1B0[4];       // GPIF will be done afer reading these
    ecc0[5] = P_ECC1B0[5];       // registers
#ifdef NAND_2K
	while(!(gpifIdle()));
#endif
    // Stored 6-byte ECC data from NAND Flash
    ecc1[0] = P_XAUTODAT1;    
    ecc1[1] = P_XAUTODAT1;    
    ecc1[2] = P_XAUTODAT1;    
    ecc1[3] = P_XAUTODAT1;    
    ecc1[4] = P_XAUTODAT1;    
    ecc1[5] = P_XAUTODAT1;    
    P_INPKTEND   = 0x84;          // skip 16-byte of the redundant data
}

//==========================================================================
//  Check ECC, if fail, return bErr=1
//  Should call CorrectData if more than ECC 2bit error, the b2BitErr will 
//  be set.
//  Modify:
//     bCnt = 0 will terminate the while loop
//     bErr = 1 will send STALL to USB Host
//==========================================================================
#ifndef NAND_2K
//#ifndef NAND_4K
void CheckECC()				//512 byte page NAND's firmware code
{
     if   ( (ecc0[0] != ecc1[0]) || (ecc0[1] != ecc1[1]) || (ecc0[2] != ecc1[2]) ||
            (ecc0[3] != ecc1[3]) || (ecc0[4] != ecc1[4]) || (ecc0[5] != ecc1[5]) )
     {
         bErr = 1; bCnt = 0; // Will stall in the Status phase
     }
}
//#endif
#endif

//==========================================================================
// Get Physical block from LUT
// return gCurZone    = (dwLBA/PageSize)/cMaxLogical
//                    = (dwLBA/PageSize) % cMaxLogical
//        gPhyAdd     = Physical block
//        gSrcAdd     = Physical Source Address for Write
//==========================================================================
void Log2Phy()
{
    BYTE page, zz, z;
    WORD wi, wTmp;
#ifdef NAND_2K
	BYTE dummy_val;
#endif

#ifdef NAND_2K
    if (bErr) return;    
    page    = xLBA3;                           // 2K: blocksize = 256 sectors
    ((BYTE *)&gDst)[0] = ((BYTE *)&dwLBA)[1];  // MSB: gDst = (WORD)(dwLBA>>16);
    ((BYTE *)&gDst)[1] = ((BYTE *)&dwLBA)[2];  // LSB:  Dst = 2K Block number

    z = 0;
    while (gDst >= cMaxLogical) { gDst -= cMaxLogical; z++; }  // z = gDst/cMaxLogical
  
    zz  = z & (gZones-1);
    z   /= gZones;                     // z = number of NAND bank

#ifdef USE_2NAND
    if (z==0) CE0_ON(), CE1_OFF(); else CE1_ON(), CE0_OFF();
#else
    gEnableBanks = IOD = nBank[z];     // enable banks
#endif
    if ((gBank != z) || (zz != gCurZone))     // need to reload when bank switch
    { 
        gBank    = z;                 // update bank
        gCurZone = zz;                // update zone


        //=================================================================
        // BuildTable(); 
        // This function build the LUT for Logical to Physical translation
        // need gCurZone
        //=================================================================
        memset16((xbyte*)gLog2Phy, MSB(cBLK_INIT), (cMaxBlock*2)/16);
        ((BYTE *)&gPhyAdd)[3] = 0;					// The same calculation stands for 4K page too as the total number of pages/sectors in a zone stays same = 64*1024
        ((BYTE *)&gPhyAdd)[2] = 0;                  // optimize for speed
        ((BYTE *)&gPhyAdd)[1] = gCurZone;           // gPhyAdd <<= 16;

        for (wi=0; wi<cMaxBlock; wi++)
        {   // read the redundant section of the page into EP6
            xbyte *p = &gLog2Phy[wi];             // get address pointer
            NandSetAdd(cNAND_READ_DATA, BadBlockMsk); //Programmable address to read the redundant/spare area
            Fifo6In();
            NandRead(cEP6, cNAND_RSIZE);
            bFreeBlk = (EP6FIFOBUF[0] == 0xff) && (EP6FIFOBUF[1] == 0xff);
            if (!bFreeBlk)
			{
                *p |= MSB(cBLK_CFG);  // special block or bad block
			}
            else
            {
                ((BYTE *)&wTmp)[0] = EP6FIFOBUF[cAddOffset];     // MSB address1
                // check if the block is not free
                if (((BYTE *)&wTmp)[0] != 0xff) 
                {
                    ((BYTE *)&wTmp)[1] = EP6FIFOBUF[cAddOffset+1];   // LSB address1
                    if ( (((BYTE *)&wTmp)[0] == EP6FIFOBUF[cAddOffset+2]) &&   // check if both addresses are matched
                         (((BYTE *)&wTmp)[1] == EP6FIFOBUF[cAddOffset+3]) )
                    {
                        xword *pTmp = &gLog2Phy[wTmp];
                        wTmp = *pTmp;
                        if (wTmp & cBLK_INIT)                   // multiple blocks map check
                        {   *pTmp =  ( wTmp & cBLK_uMSK ) | wi; // store the remap block                    
                            *p    |= MSB(cBLK_USE);             // Set Block Use
                        }
                        else
                        {
							NandSendCmd(cNAND_RESET); _nop_();_nop_();				
                            nand_blk_erase(gPhyAdd);
							nand_get(dummy_val);
							if (dummy_val&0x01)		//if the erase fails, the block is bad
							{
								*p |= MSB(cBLK_BAD);
								write0ForBadBlock();
							}
                        }
                    }
                }  // else this block is free
            }
            gPhyAdd += c2KPageSize;  // next block for 2K NAND, this size is same in 4K NAND = 64
        } // for wi
        bFreeFound = 0;               // cross the Zone, reset the Free Block Found
    } // update table again

    wTmp  = zz << cMaxBlock2N;        // compute Zone 
    pDst  = gLog2Phy + gDst;          // pDst = &gLog2Phy[gDst]
    gSrc  = *pDst & cBLK_aMSK;        // gLog2Phy[gDst]: Get Physical block

    if (directionIn)                  // read just get the table
    {
        wTmp |= gSrc;    
        ((BYTE *)&gPhyAdd)[3] = ( ((BYTE *)&wTmp)[1] << 6 ) | (page>>2);
        ((BYTE *)&gPhyAdd)[2] = ( ((BYTE *)&wTmp)[1] >> 2 ) | ( ((BYTE *)&wTmp)[0] << 6 );
        ((BYTE *)&gPhyAdd)[1] = ( ((BYTE *)&wTmp)[0] >> 2 );

		NandSetAdd(cNAND_READ_DATA, (page&SectorbCntMsk));
    }
    else
    {
        xbyte  *p;
        nGetFreeBlk();                                // get Free Block
        wi = gFreeBlk;
        if (gSrc&cBLK_INIT) gSrc=wi;                  // default value set to free block
        bNeedErase = (wi != gSrc);                    // don't erase if Src = Dst && block free

        // *(xbyte*)&gLog2Phy[gSrc] &=  MSB(~cBLK_USE);  // set this block free
        p = (xbyte*)&gLog2Phy[gSrc];                  // point to the SRC block
        // July-07-05 Free Block must exclude CFG block as well
        bFreeBlk = !(*p & MSB(cBLK_uMSK));             // free block 
        *p = *p & MSB(~cBLK_USE);                      // set this block free
        
        *(xbyte*)&gLog2Phy[wi]   |=  MSB(cBLK_USE);   // Set this block is used        
        *pDst = (*pDst & cBLK_uMSK) | wi;             // update Logical address

        wi |= wTmp;    
        ((BYTE *)&gPhyAdd)[3] = ( ((BYTE *)&wi)[1] << 6 );
        ((BYTE *)&gPhyAdd)[2] = ( ((BYTE *)&wi)[1] >> 2 ) | ( ((BYTE *)&wi)[0] << 6 );
        ((BYTE *)&gPhyAdd)[1] = ( ((BYTE *)&wi)[0] >> 2 );

        wTmp |= gSrc;    
        gSrcBlk0 = ((BYTE *)&gSrcAdd)[3] = ( ((BYTE *)&wTmp)[1] << 6 );
        ((BYTE *)&gSrcAdd)[2] = ( ((BYTE *)&wTmp)[1] >> 2 ) | ( ((BYTE *)&wTmp)[0] << 6 );
        ((BYTE *)&gSrcAdd)[1] = ( ((BYTE *)&wTmp)[0] >> 2 );

        if (page) NAND_PCPY(page, 0);
    }
#else	//512 byte code -------------------------------------------------------------------------
    BYTE npage;

    if (bErr) return;    
    gSrc    = (WORD)dwLBA; 
    ((BYTE *)&gDst)[0] = ((BYTE *)&dwLBA)[0];  // MSB: gDst = (WORD)(dwLBA>>16);
    ((BYTE *)&gDst)[1] = ((BYTE *)&dwLBA)[1];  // LSB 
    if (bInterLeave)
    {    
        page = xLBA3 & cInterLeaveMsk;      
        npage = page>>1;                       // remove bit0 for interleave
        gDst <<= 16-(c512PageSize2N+1);        // dwLBA >>= c512PageSize2N+1
        gSrc >>= c512PageSize2N+1;
    }
    else
    {
        npage = page = xLBA3 & (c512PageSize-1); // mask page 5-bits 0x1F
        gDst <<= 16-c512PageSize2N;              // dwLBA >>= c512PageSize2N
        gSrc >>= c512PageSize2N;
    }   
    gDst |= gSrc;                                // gDst = dwLBA/c512PageSize

    z = 0;
    while (gDst >= cMaxLogical) { gDst -= cMaxLogical; z++; }  // z = gDst/cMaxLogical   
    zz  =  z & (gZones-1);    // zz = zone numbers within a NAND
    z   /= gZones;            // gZones=8=8192, gZones=4=4096

#ifdef USE_2NAND
    CE0_ON(),CE1_ON();       // enable both NAND chip select
#else
    gEnableBanks = IOD = aBanks[z];  // pre-load fast variables
#endif
  
    //  Reload the LUT when cross bank and zone
    if ( zz != gCurZone || z != gBank)
    { 
#ifndef USE_2NAND
        gEnableBank0 = aBank0[z];        // pre-load fast variables
        gEnableBank1 = aBank1[z];        // pre-load fast variables
#endif
        gBank    = z;                    // update bank
        gCurZone = zz;
        // 1K Zone
        // gPhyAdd  = ((DWORD)gCurZone << (cMaxBlock2N+c512PageSize2N));   // Get Page address
        ((BYTE *)&gPhyAdd)[3] = 0;
        ((BYTE *)&gPhyAdd)[2] = gCurZone<<7; 
        ((BYTE *)&gPhyAdd)[1] = gCurZone>>1;

        // 2K Zone
        //((BYTE *)&gPhyAdd)[3] = 0;
        //((BYTE *)&gPhyAdd)[2] = 0;                                        // optimize for speed
        //((BYTE *)&gPhyAdd)[1] = gCurZone;

        //==========================================================================
        // This function build the LUT for Logical to Physical translation
        // need gCurZone
        //==========================================================================

        memset16((xbyte*)gLog2Phy, MSB(cBLK_INIT), (cMaxBlock*2)/16);
        for (wi=0; wi<cMaxBlock; wi++)
        {   // read the redundant section of the page into EP6           
            NandSetAdd(cNAND_READ_REDUNDANT, 0); 
            for (z=0; z <3; z++) // retry 3 times
            {
                xbyte *p = &gLog2Phy[wi];             // get address pointer
                // for InterLeave case
                // if either block in bank0 or bank1 is bad, will mark both blocks bad
                if (bInterLeave) bank_select(cBank0);  // check bank0                
                Fifo6In();
                NandRead(cEP6, 16);
                bFreeBlk = (EP6FIFOBUF[0] == 0xff) && (EP6FIFOBUF[1] == 0xff);
                if (bInterLeave)                       // check bank1
                { 
                    bank_select(cBank1);
                    NandRead(cEP6, 16);
                    if (EP6FIFOBUF[16] != 0xff || EP6FIFOBUF[17] != 0xff) bFreeBlk = 0;
                }
                if (!bFreeBlk) 
                {
                    *p |= MSB(cBLK_CFG);  // special block or bad block
                    break;
                }
                else
                {
                    ((BYTE *)&wTmp)[0] = EP6FIFOBUF[cAddOffset];     // MSB address1
                    if  (((BYTE *)&wTmp)[0]==0xff) break;            // free block
                    ((BYTE *)&wTmp)[1] = EP6FIFOBUF[cAddOffset+1];   // LSB address1

                    if ( (((BYTE *)&wTmp)[0] == EP6FIFOBUF[cAddOffset+2]) &&   // check if both addresses are matched
                         (((BYTE *)&wTmp)[1] == EP6FIFOBUF[cAddOffset+3]) )
                    {
                        xword *pTmp = &gLog2Phy[wTmp];
                        wTmp = *pTmp;
                        if (wTmp & cBLK_INIT)                   // multiple blocks map check
                        {
                            *pTmp =  ( wTmp & cBLK_uMSK ) | wi; // store the remap block                    
                            *p    |= MSB(cBLK_USE);             // Set Block Use
                        }
                        else
                        {
                            NandSendCmd(cNAND_RESET);           // fix the multiple blocks case
                            nand_blk_erase(gPhyAdd); 
                        }
                        break;
                    }
                }
                if (z == 2) *p |= MSB(cBLK_BAD);
            }
            gPhyAdd += c512PageSize;  // next block
        } // for gi
        bank_default();               // set default
        NandSendCmd(cNAND_RESET);     // must reset afer REDUNDANT read
        bFreeFound = 0;               // cross the Zone, reset the Free Block Found
    } // update table again

    wTmp  = zz << cMaxBlock2N;        // compute Zone 
    pDst  = gLog2Phy + gDst;          // pDst = &gLog2Phy[gDst]
    gSrc  = *pDst & cBLK_aMSK;        // gLog2Phy[gDst]: Get Physical block
    if (directionIn)                  // Host Read data
    {
        wTmp |= gSrc;    
        nand_ready3();
        NAND_CLE  = 1;
        P_XGPIFSGLDATLX = cNAND_READ_DATA;
        NAND_CLE  = 0; 
        NAND_ALE  = 1;
        P_XGPIFSGLDATLX = 0;     
        P_XGPIFSGLDATLX = ( ((BYTE *)&wTmp)[1] << 5 ) | npage;
        P_XGPIFSGLDATLX = ( ((BYTE *)&wTmp)[1] >> 3 ) | ( ((BYTE *)&wTmp)[0] << 5 );
        P_XGPIFSGLDATLX = ( ((BYTE *)&wTmp)[0] >> 3 );
        NAND_ALE  = 0;              
        Fifo6In();
        if (bInterLeave) ready_ignore(); 
        nand_ready3();       
    }
    else
    {   
        nGetFreeBlk();                                // get Free Block
        wi = gFreeBlk;
        if (gSrc&cBLK_INIT) gSrc=wi;                  // default value set to free block
        bNeedErase = (wi != gSrc);                    // don't erase if Src = Dst && block free
        *(xbyte*)&gLog2Phy[gSrc] &=  MSB(~cBLK_USE);  // set this block free
        *(xbyte*)&gLog2Phy[wi]   |=  MSB(cBLK_USE);   // Set this block is used        
        *pDst = (*pDst & cBLK_uMSK) | wi;             // update Logical address
 
        wi |= wTmp;    
        //gPhyAdd = ((DWORD)(wTmp | wi) << c512PageSize2N);
        ((BYTE *)&gPhyAdd)[3] = ( ((BYTE *)&wi)[1] << 5 );
        ((BYTE *)&gPhyAdd)[2] = ( ((BYTE *)&wi)[1] >> 3 ) | ( ((BYTE *)&wi)[0] << 5 );
        ((BYTE *)&gPhyAdd)[1] = ( ((BYTE *)&wi)[0] >> 3 );
        //gSrcAdd = ((DWORD)(wTmp | gSrc) << c512PageSize2N);
        wTmp |= gSrc;    
        gSrcBlk0 = ((BYTE *)&gSrcAdd)[3] = ( ((BYTE *)&wTmp)[1] << 5 );
        ((BYTE *)&gSrcAdd)[2] = ( ((BYTE *)&wTmp)[1] >> 3 ) | ( ((BYTE *)&wTmp)[0] << 5 );
        ((BYTE *)&gSrcAdd)[1] = ( ((BYTE *)&wTmp)[0] >> 3 );
        // check for head copy here
        if (page) nCopyPages(page, 0);
    }
#endif
    DBUG_DUMP();
}

//==========================================================================
// Nand Copy Pages:
//   always perform ECC correction for SOFT error.
//   If more than 2-bit error, return b2BitErr will be set
//==========================================================================
#ifdef NAND_4K			
// nCopyPages is special for 4K as it takes the whole sector (1024 bytes) into one endpoint, checks error and copies it all back for improved performance
// hence it requires two extra variables ecc2[] and ecc3[]
void nCopyPages(BYTE cnt, BYTE cc)
{
    BYTE msk = cc & SectorbCntMsk;
	BYTE dummy;
	xbyte Stat0, Stat1, Stat2, Stat3;

    do
    { 
        Fifo6In();
        P_GPIFTCB1 = MSB(cNAND_DSIZE);               	// setup GPIF Count
        P_ECCRESET = P_GPIFTCB0 = LSB(cNAND_DSIZE);  	// Reading first 512, 2(status)+6(ECC) bytes will be saved directly into variables
        n2k_radd(gSrcAdd, msk);
        GPIFTRIG = 0x04 | cEP6;                       // Arm EP6

        bCnt = --cnt;

        // When copy pages, all the data will have ECC correction
        while (P_GPIFTCB1);          // Read the first 256 ECC registers
        ecc0[0] = P_ECC1B0[0];       // save to local buffers
        ecc0[1] = P_ECC1B0[1]; 
        ecc0[2] = P_ECC1B0[2]; 
        while (!gpifIdle());
        //while (P_GPIFTCB0 & 0xf8);   // Read second 256 ECC registers       
        ecc0[3] = P_ECC1B0[3];       // save ECC 
        ecc0[4] = P_ECC1B0[4];       // GPIF will be done afer reading these
        ecc0[5] = P_ECC1B0[5];       // registers

		P_GPIFTCB1 = MSB(cNAND_DSIZE);					// Setup GPIF to read the second half which is stored back to back on the Flash
		P_ECCRESET = P_GPIFTCB0 = LSB(cNAND_DSIZE);
		GPIFTRIG = 0x04 | cEP6;							// Arm EP6 to read next 512 bytes

		_nop_();				// Required breather for GPIF engine, removal causes synchronization issues

		while (P_GPIFTCB1);			 // Read the first 256 ECC regs for second half of the sector
		ecc2[0] = P_ECC1B0[0];		 // Saving to local buffers
		ecc2[1] = P_ECC1B0[1];
		ecc2[2] = P_ECC1B0[2];
		while (!gpifIdle());
		ecc2[3] = P_ECC1B0[3];
		ecc2[4] = P_ECC1B0[4];
		ecc2[5] = P_ECC1B0[5];

		// Read Spare area (16 bytes) for first half of the sector
		dummy = P_XGPIFSGLDATLX;				// getting rid of the buffered unwanted value
		Stat0 = P_XGPIFSGLDATLX;// _nop_();
		Stat1 = P_XGPIFSGLDATLX;// _nop_();
        ecc1[0] = P_XGPIFSGLDATLX;
        ecc1[1] = P_XGPIFSGLDATLX;
        ecc1[2] = P_XGPIFSGLDATLX;
        ecc1[3] = P_XGPIFSGLDATLX;
        ecc1[4] = P_XGPIFSGLDATLX;
        ecc1[5] = P_XGPIFSGLDATLX;
		dummy = P_XGPIFSGLDATLX;
		dummy = P_XGPIFSGLDATLX;
		dummy = P_XGPIFSGLDATLX;
		dummy = P_XGPIFSGLDATLX;
		dummy = P_XGPIFSGLDATLX;
		dummy = P_XGPIFSGLDATLX;
		dummy = P_XGPIFSGLDATLX;
		dummy = P_XGPIFSGLDATLX;

		// Read Spare area (16 bytes) for second half of the sector
		Stat2 = P_XGPIFSGLDATLX;// _nop_();
		Stat3 = P_XGPIFSGLDATLX;// _nop_();
        ecc3[0] = P_XGPIFSGLDATLX;
        ecc3[1] = P_XGPIFSGLDATLX;
        ecc3[2] = P_XGPIFSGLDATLX;
        ecc3[3] = P_XGPIFSGLDATLX;
        ecc3[4] = P_XGPIFSGLDATLX;
        ecc3[5] = P_XGPIFSGLDATLX;
		dummy = P_XGPIFSGLDATLX;
		dummy = P_XGPIFSGLDATLX;
		dummy = P_XGPIFSGLDATLX;
		dummy = P_XGPIFSGLDATLX;
		dummy = P_XGPIFSGLDATLX;
		dummy = P_XGPIFSGLDATLX;
		dummy = P_XGPIFSGLDATLX;
		dummy = P_XGPIFSGLDATLX;

		CorrectData();
		CorrectDataX();

        fifo6_out();                        // CFG Out for copy data  
        n2k_wadd(gPhyAdd, msk);
        P_GPIFTCB1 = MSB(1023);    // setup GPIF Count to write the complete sector
        P_GPIFTCB0 = LSB(1023);    
        GPIFTRIG=cEP6;

        cc++;
        msk = cc & 3;                       // use by 2K NAND
        if (msk==0) { xSrcAdd++; xPhyAdd++; }      
        while (!gpifIdle());				

		// Write Spare area (16 bytes) for first half of the sector
		P_XGPIFSGLDATLX = Stat0;     _nop_();
		P_XGPIFSGLDATLX = Stat1;     _nop_();
		P_XGPIFSGLDATLX = ecc1[0];   _nop_();
		P_XGPIFSGLDATLX = ecc1[1];   _nop_();
		P_XGPIFSGLDATLX = ecc1[2];   _nop_();
		P_XGPIFSGLDATLX = ecc1[3];   _nop_();
		P_XGPIFSGLDATLX = ecc1[4];   _nop_();
		P_XGPIFSGLDATLX = ecc1[5];   _nop_();
		WriteSpare();

        // Write Spare area (16 bytes) for second half of the sector
		P_XGPIFSGLDATLX = Stat2;     _nop_();
		P_XGPIFSGLDATLX = Stat3;     _nop_();
		P_XGPIFSGLDATLX = ecc3[0];   _nop_();
		P_XGPIFSGLDATLX = ecc3[1];   _nop_();
		P_XGPIFSGLDATLX = ecc3[2];   _nop_();
		P_XGPIFSGLDATLX = ecc3[3];   _nop_();
		P_XGPIFSGLDATLX = ecc3[4];   _nop_();
		P_XGPIFSGLDATLX = ecc3[5];   _nop_();
		WriteSpare();

		nand_send_command(cNAND_PROGRAM_PAGE);_nop_();_nop_();				
    } while ( bCnt ); 
}

#else				//Code for 2K and 512B

void nCopyPages(BYTE cnt, BYTE cc)
{
#ifdef NAND_2K
    BYTE msk=cc&3;   
    do
    { 
        Fifo6In();
        P_GPIFTCB1 = MSB(cNAND_DSIZE+8);               // setup GPIF Count
        P_ECCRESET = P_GPIFTCB0 = LSB(cNAND_DSIZE+8);  // Reading 512+2(status)+6(ECC)
        n2k_radd(gSrcAdd, msk);
        GPIFTRIG = 0x04 | cEP6;                       // Arm EP6
        bCnt = --cnt;   
        // When copy pages, all the data will have ECC correction
        AUTOPTR1H = MSB(cEP6FIFO);
        AUTOPTR1L = LSB(cEP6FIFO);
        while (P_GPIFTCB1);          // Read the first 256 ECC registers
        ecc0[0] = P_ECC1B0[0];       // save to local buffers
        ecc0[1] = P_ECC1B0[1]; 
        ecc0[2] = P_ECC1B0[2]; 
        while (!gpifIdle());
        //while (P_GPIFTCB0 & 0xf8);   // Read second 256 ECC registers       
        ecc0[3] = P_ECC1B0[3];       // save ECC 
        ecc0[4] = P_ECC1B0[4];       // GPIF will be done afer reading these
        ecc0[5] = P_ECC1B0[5];       // registers
        // Stored 6-byte ECC data from NAND Flash
        ecc1[0] = P_XAUTODAT1;    
        ecc1[1] = P_XAUTODAT1;    
        ecc1[2] = P_XAUTODAT1;    
        ecc1[3] = P_XAUTODAT1;     
        ecc1[4] = P_XAUTODAT1;    
        ecc1[5] = P_XAUTODAT1;    
        CorrectData();
        fifo6_out();                        // CFG Out for copy data  
        n2k_wadd(gPhyAdd, msk);
        P_GPIFTCB1 = MSB(cNAND_DSIZE+8);    // setup GPIF Count: Ignore 4-bytes
        P_GPIFTCB0 = LSB(cNAND_DSIZE+8);    // Reading 528 take 36us
        GPIFTRIG=cEP6;
        cc++;
        msk = cc & 3;                       // use by 2K NAND
        if (msk==0) { xSrcAdd++; xPhyAdd++; }      
        while (!gpifIdle());
		WriteSpare();
        nand_send_command(cNAND_PROGRAM_PAGE);_nop_();_nop_();				
    } while ( bCnt ); 
#else	//512 byte code -----------------------------------------------------------------
    bit lba0 = cc&1;
    do
    { 
        Fifo6In();
        P_GPIFTCB1 = MSB(cNAND_DSIZE+8);               // setup GPIF Count
        P_ECCRESET = P_GPIFTCB0 = LSB(cNAND_DSIZE+8);  // Reading 512+2(status)+6(ECC)
        if (bInterLeave)
        {
#ifdef USE_2NAND
            if (lba0) CE0_OFF(), CE1_ON(); else CE0_ON(), CE1_OFF();
#else
            if (lba0) IOD=gEnableBank1; else IOD=gEnableBank0;
#endif
        }
        nand_ready3();                            // wait for previous program done
        n512_setadd(cNAND_READ_DATA,gSrcAdd);     // read program pages            
        GPIFTRIG = 0x04 | cEP6;                   // Arm EP6
        lba0 = !lba0;          
        bCnt = --cnt;   

        // When copy pages, all the data will have ECC correction
        AUTOPTR1H = MSB(cEP6FIFO);
        AUTOPTR1L = LSB(cEP6FIFO);
        while (P_GPIFTCB1);          // Read the first 256 ECC registers
        ecc0[0] = P_ECC1B0[0];       // save to local buffers
        ecc0[1] = P_ECC1B0[1]; 
        ecc0[2] = P_ECC1B0[2]; 

        while (P_GPIFTCB0 & 0xf8);   // Read second 256 ECC registers       
        ecc0[3] = P_ECC1B0[3];       // save ECC 
        ecc0[4] = P_ECC1B0[4];       // GPIF will be done afer reading these
        ecc0[5] = P_ECC1B0[5];       // registers
        // Stored 6-byte ECC data from NAND Flash
        ecc1[0] = P_XAUTODAT1;    
        ecc1[1] = P_XAUTODAT1;    
        ecc1[2] = P_XAUTODAT1;    
        ecc1[3] = P_XAUTODAT1;     
        ecc1[4] = P_XAUTODAT1;    
        ecc1[5] = P_XAUTODAT1;    
        CorrectData();
        fifo6_out();                         // CFG Out for copy data            
        P_GPIFTCB1 = MSB(cNAND_DSIZE+8);     // setup GPIF Count: Ignore 4-bytes
        P_GPIFTCB0 = LSB(cNAND_DSIZE+8);     // Reading 528 take 36us
        n512_setadd(cNAND_WRITE_DATA, gPhyAdd);
        GPIFTRIG=cEP6;
        if (!lba0 || bInterLeave==0) { xSrcAdd++; xPhyAdd++; }      
        while (!gpifIdle());
        P_XGPIFSGLDATLX = MSB(gDst);               // Phy2Log
        P_XGPIFSGLDATLX = LSB(gDst);        
        P_XGPIFSGLDATLX = MSB(gDst);        
        P_XGPIFSGLDATLX = LSB(gDst);        
        nand_send_command(cNAND_PROGRAM_PAGE);//_nop_();_nop_();				
    } while ( bCnt ); 
#endif

}
#endif

//==========================================================================
//  Adding this code to support Wear-Leveling. 
//  Note: This subroutine should call only USB Idle, NAND Busy or
//  Waiting GPIF Idle.
//==========================================================================
void nSearchFreeBlock(BYTE i)
{
    WORD tmp;
    if (bFreeFound) return;      // if found do need to search again
    tmp = (WORD)gLog2Phy + gFreeBlk*2;
    AUTOPTR1H = MSB(tmp); AUTOPTR1L = LSB(tmp);
    do
    {
        if (!(P_XAUTODAT1 & MSB(cBLK_uMSK)))
        {
            bFreeFound = 1;            // Mark found a free erased block
            return;
        }
        P_XAUTODAT1;
        gFreeBlk++;                // Search Next block
        if (gFreeBlk & cMaxBlock)
            gFreeBlk = 0, AUTOPTR1H = MSB(gLog2Phy), AUTOPTR1L = LSB(gLog2Phy);
    } while (--i);                     // 1-256 search entries
}

//==========================================================================
// Write NAND Pages: support both interleave and non-interleave code
//   Use gFreeBlk as the back ground searching free erased block
//==========================================================================
void nWritePages()
{

	
#ifdef NAND_2K
    bit prog,t;
    BYTE wcmd, dummy_read, xLBA3s;
	bErr = 0;

    if (gPartialCpy==0) Log2Phy();  

    // Need to queue up more buffers to avoid hang up on the Write transfer
    // for 512-byte just wait for empty
    if ((gSectorcount>1) || !EZUSB_HIGHSPEED())
       while (!(P_EP2CS & bmEPFULL));     // wait for host send all data
    else
#ifdef NAND_4K
	   while (!(P_EP2CS & bmEPFULL));	  // in case of 1024 byte sector, wait for host to write whole sector
#else
       while ((P_EP2CS & bmEPEMPTY));     // Wait for host to send data
#endif
    ARM_EP2();
    P_EP2FIFOCFG = bmAUTOOUT | bmOEP;    // set auto mode

	NandSetAdd(cNAND_WRITE_DATA, (xLBA3 & SectorbCntMsk));		
    do 
    {
//#ifdef NAND_4K											// It is expected that the host would have sent next packet by the time we process the present; if found otherwise, please use this wait
  		//while(!(P_EP2CS & bmEPFULL));					// wait for EP2 to get filled every sector - removed for optimization
//#endif
        P_GPIFTCB1 = MSB(cNAND_DSIZE);
        P_ECCRESET = P_GPIFTCB0 = LSB(cNAND_DSIZE);
        GPIFTRIG = cEP2;

        //-----------------------------------------------------
        // PreProcess calculation while waiting transfer done
        //-----------------------------------------------------
        bCnt = --gSectorcount;
		xLBA3s = xLBA3;

#ifdef NAND_4K											// A sector in 4K page is of size 1024 bytes, followed by 32 bytes of ECC
		while (!gpifIdle());							// waiting till the transaction is over
		ecc2[0] = P_ECC1B0[0];							// Saving the ECC bits for redundant/spare area write
		ecc2[1] = P_ECC1B0[1];
		ecc2[2] = P_ECC1B0[2];
		ecc2[3] = P_ECC1B0[3];
		ecc2[4] = P_ECC1B0[4];
		ecc2[5] = P_ECC1B0[5];
		P_GPIFTCB1 = MSB(cNAND_DSIZE);					// setting up transfer for the second half of the sector
		P_ECCRESET = P_GPIFTCB0 = LSB(cNAND_DSIZE);		// reset the ECC engine to capture new ECC bits
		GPIFTRIG = cEP2;
#endif

        dwLBA++;                                // next sector
        prog = (xLBA3 & SectorbCntMsk)==0;
        bReload = (xLBA3 == 0);

        if( bProgCache )
           wcmd = cNAND_PROGRAM_CACHE;
        else
           wcmd = cNAND_PROGRAM_PAGE;

        if (bReload || !bCnt) wcmd = cNAND_PROGRAM_PAGE; 
        
        bReload &= bCnt;
		t=prog;
        if (prog) { xPhyAdd++; xSrcAdd++;  }    // next address       
        prog |= !bCnt;                   
        while (!gpifIdle());
#ifdef NAND_4K									// redundant write for the first half of the sector
	    P_XGPIFSGLDATLX = 0xff;		_nop_();
   		P_XGPIFSGLDATLX = 0xff;		_nop_();
   		P_XGPIFSGLDATLX = ecc2[0];	_nop_();
   		P_XGPIFSGLDATLX = ecc2[1];	_nop_();
   		P_XGPIFSGLDATLX = ecc2[2];	_nop_();
   		P_XGPIFSGLDATLX = ecc2[3];	_nop_();
   		P_XGPIFSGLDATLX = ecc2[4];	_nop_();
   		P_XGPIFSGLDATLX = ecc2[5];	_nop_();
   		P_XGPIFSGLDATLX = MSB(gDst);_nop_();
   		P_XGPIFSGLDATLX = LSB(gDst);_nop_();
   		P_XGPIFSGLDATLX = MSB(gDst);_nop_();
   		P_XGPIFSGLDATLX = LSB(gDst);_nop_();
   		P_XGPIFSGLDATLX = 0xff;		_nop_();
   		P_XGPIFSGLDATLX = 0xff;		_nop_();
   		P_XGPIFSGLDATLX = 0xff;		_nop_();
   		P_XGPIFSGLDATLX = 0xff;		_nop_();
#endif
        WriteRedundant();						// This #define writes spare area with ecc from P_ECC1B0 register, hence it works fine with 2K page and 4K page NANDS
        if (prog) 
        {
            nand_send_command(wcmd); _nop_();_nop_();				
			nand_ready3();
			NandSendCmd(cNAND_READ_STATUS);		// removed nop from the dummy reads for optimization purposes
			dummy_read = P_XGPIFSGLDATLX;		// Check if the NAND notifies a bad status after write
			dummy_read = P_XGPIFSGLDATLNOX;
			if(dummy_read&0x01) 
			{
				bErr = 1;
				bCnt = 0;
			} 
			else 								// Even if the write succeeded without bad status reporting back, verify the start of spare area written to NAND
			{									// this functionality can be removed if performance is an issue
				if(t){xPhyAdd--;}				// t preserves the LBA address at the page boundaries for recheck
				NandSetAdd(cNAND_READ_DATA, (xLBA3s&SectorbCntMsk)+SpareOffset);  //This is to check if the block is bad and was missed by incorrect status read
            	Fifo6In();
            	NandRead(cEP6, 3);				// changed from RSIZE=12bytes to 4bytes for optimization purposes
				if (EP6FIFOBUF[0]!=0xFF || EP6FIFOBUF[1]==0x00)
				{
					bErr = 1;
					bCnt = 0;
				}
				else
					bErr = 0;
				if(t){xPhyAdd++;}
			}

			if (bReload&&!bErr) { 
								nEraseBlock();
								Log2Phy();
								}
            if (bCnt) n2k_set_wadd(); 
        }        
    } while (bCnt);

    gPartialCpy = xLBA3;			

    // partial block must leave the nand chip select asserted
    if (gPartialCpy)					// gPartialCpy needs to hold max value = 255
        gPartialCpy = 0-gPartialCpy;
    else  
    {
        nEraseBlock();
        DISABLE_NAND();
    }
    P_EP2FIFOCFG = 0;
#else		// 512 byte code---------------------------------------------------
    if (gPartialCpy==0) Log2Phy();  
    // Need to queue up more buffers to avoid hang up on the Write transfer
    // for 512-byte just wait for empty
    if ((gSectorcount>1) || !EZUSB_HIGHSPEED())
       while (!(P_EP2CS & bmEPFULL));     // wait for host send all data
    else
       while ((P_EP2CS & bmEPEMPTY));     // Wait for host to send data
    ARM_EP2(); 
    P_EP2FIFOCFG = bmAUTOOUT | bmOEP;     // set auto mode
    do 
    {
        P_GPIFTCB1 = MSB(cNAND_DSIZE);
        P_ECCRESET = P_GPIFTCB0 = LSB(cNAND_DSIZE);
        if (bInterLeave)
        {
#ifdef USE_2NAND
            if (bLBA0) {  CE1_ON(), CE0_OFF(); nand_ready1(); }     // bank1
            else       {  CE0_ON(), CE1_OFF(); nand_ready0(); }     // bank0
#else
            if (bLBA0) {  IOD=gEnableBank1; nand_ready1(); }     // bank1
            else       {  IOD=gEnableBank0; nand_ready0(); }     // bank0
#endif
        }
        else nand_ready3();    
        n512_setadd(cNAND_WRITE_DATA,gPhyAdd);                   // send program pages             
        GPIFTRIG = cEP2;
        //-----------------------------------------------------
        // PreProcess calculation while waiting transfer done
        //-----------------------------------------------------
        bCnt = --gSectorcount;
        dwLBA++;                                         // next sector
        if (bInterLeave)
           gPartialCpy = xLBA3 & cInterLeaveMsk;
        else
           gPartialCpy = xLBA3 & (c512PageSize-1);
        if (!bLBA0 || bInterLeave==0) { xPhyAdd++; xSrcAdd++;  }       // next address
        
        bReload = bCnt && gPartialCpy==0;
        while (!gpifIdle());
        WriteRedundant();
        nand_send_command(cNAND_PROGRAM_PAGE);//_nop_();_nop_();			
        // nSearchFreeBlock(64); 
        if (bReload) { nEraseBlock(); Log2Phy(); }
    } while (bCnt);

    // partial block must leave the nand chip select asserted
    if (gPartialCpy)
    { 
        if (bInterLeave) gPartialCpy = 64-gPartialCpy; 
        else gPartialCpy = c512PageSize-gPartialCpy;
    } 
    else  
    {
        nEraseBlock();
        DISABLE_NAND();
    }
    P_EP2FIFOCFG = 0;
#endif
}

//==========================================================================
// Redundant mapping for Configuration block
//  0-9: 0xff, 01, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff
//  a-f: ecc1b0, ecc1b1, ecc1b2, ecc2b0, ecc2b1, ecc2b2   
// Redundant mapping for data block
//  0-7: 0xff, 0xff, ecc1b0, ecc1b1, ecc1b2, ecc2b0, ecc2b1, ecc2b2
//  8-b: add1, add2
//  c-f: ignore
// set b2BitErr when more than 2 bit ECC error
//==========================================================================
bit CorrectData()
{    
    BYTE a,b,c,x,i,t;

    for (i=0;  i<2;  i++)
    {       
		t = i*3;
        b = ecc0[0+t] ^ ecc1[0+t];
        a = ecc0[1+t] ^ ecc1[1+t];
        c = ecc0[2+t] ^ ecc1[2+t];
        if (a|b|c)
        {
            if ((((a^(a>>1)) & 0x55) == 0x55) && // Does each pair of parity bits contain
                (((b^(b>>1)) & 0x55) == 0x55) && // one error and one match?
                 (((c^(c>>1)) & 0x54) == 0x54))
            {
                x = (a & 0x80); // If so, there's a one-bit error in data[].
                if (a & 0x20) x |= 0x40; // Find which byte is in error...
                if (a & 0x08) x |= 0x20;
                if (a & 0x02) x |= 0x10;
                if (b & 0x80) x |= 0x08;
                if (b & 0x20) x |= 0x04;
                if (b & 0x08) x |= 0x02;
                if (b & 0x02) x |= 0x01;
                a = 0; // ... and which bit...
                if (c & 0x80) a |= 0x04; 
                if (c & 0x20) a |= 0x02;
                if (c & 0x08) a |= 0x01;
                EP6FIFOBUF[x+ (WORD)(i<<8)] ^= (1 << a); // ... and correct it.
                // one bit is ok
            }
            else
			{
				b2BitErr=1;
				return cFail;
			}  //  two bits return error
        }
    }
    return cOK;
}

#ifdef NAND_4K

bit CorrectDataX()					// ECC check/correct data for the second half of the sector, required only in 4K firmware
{    
    BYTE a,b,c,x,i,t;

    for (i=0;  i<2;  i++)
    {       
		t = i*3;
        b = ecc2[0+t] ^ ecc3[0+t];
        a = ecc2[1+t] ^ ecc3[1+t];
        c = ecc2[2+t] ^ ecc3[2+t];
        if (a|b|c)
        {
            if ((((a^(a>>1)) & 0x55) == 0x55) && // Does each pair of parity bits contain
                (((b^(b>>1)) & 0x55) == 0x55) && // one error and one match?
                 (((c^(c>>1)) & 0x54) == 0x54))
            {
                x = (a & 0x80); // If so, there's a one-bit error in data[].
                if (a & 0x20) x |= 0x40; // Find which byte is in error...
                if (a & 0x08) x |= 0x20;
                if (a & 0x02) x |= 0x10;
                if (b & 0x80) x |= 0x08;
                if (b & 0x20) x |= 0x04;
                if (b & 0x08) x |= 0x02;
                if (b & 0x02) x |= 0x01;
                a = 0; // ... and which bit...
                if (c & 0x80) a |= 0x04; 
                if (c & 0x20) a |= 0x02;
                if (c & 0x08) a |= 0x01;
                EP6FIFOBUF[x+ (WORD)(i<<8) + 512] ^= (1 << a); // ... and correct it.
                // one bit is ok
            }
            else
			{
				b2BitErr=1;
				return cFail;
			}  //  two bits return error
        }
    }
    return cOK;
}

#endif

#ifdef NAND_2K

//==========================================================================
// Internal Move:
//    Partial page will use nCopyPages, the full 2K page will use 
//    internal move
//==========================================================================
void nNandMove(BYTE cnt, BYTE cc)
{
#ifdef NAND_4K
	nCopyPages(cnt,cc);
	return;
#else
    BYTE z, msk = cc&3;

    if (!bInternalMove) { nCopyPages(cnt,cc); return; }

    if (msk) { z = 4-msk; nCopyPages(z, msk); cnt -= z; }
    msk = cnt/4;

    // if the src is blank, need to have valid page
    if (cc==0 && msk && bFreeBlk)
    {
       nCopyPages(4, 0); // ONLY need to copy 1st page
       cnt -= 4;
       msk--;
    }           
    for (z=0; z<msk; z++)
    {              
        nand_ready3();
        NAND_CLE=1,P_XGPIFSGLDATLX=0, NAND_CLE=0;
        NAND_ALE = 1;
        P_XGPIFSGLDATLX = 0; cnt -= 4; P_XGPIFSGLDATLX = 0;
        P_XGPIFSGLDATLX = ((BYTE *)&gSrcAdd)[3]; 
        P_XGPIFSGLDATLX = ((BYTE *)&gSrcAdd)[2];    
        P_XGPIFSGLDATLX = ((BYTE *)&gSrcAdd)[1];     
        NAND_ALE = 0;                       
        NAND_CLE=1, P_XGPIFSGLDATLX=0x35, NAND_CLE=0;
        
        nand_ready3();  // wait for tr
        NAND_CLE=1,P_XGPIFSGLDATLX=0x85, NAND_CLE=0;
        NAND_ALE = 1;
        P_XGPIFSGLDATLX = 0; xSrcAdd++; P_XGPIFSGLDATLX = 0;
        P_XGPIFSGLDATLX = ((BYTE *)&gPhyAdd)[3]; 
        P_XGPIFSGLDATLX = ((BYTE *)&gPhyAdd)[2];    
        P_XGPIFSGLDATLX = ((BYTE *)&gPhyAdd)[1];     
        NAND_ALE = 0;                       
        NAND_CLE=1, P_XGPIFSGLDATLX=0x10, NAND_CLE=0;
        xPhyAdd++;
    }  
    if (cnt) nCopyPages(cnt, 0);
#endif
}

#endif
