//-----------------------------------------------------------------------------
// Copyright (c) 2005 Cypress Semiconductor, Inc. All rights reserved
//
// File:       inand.h
// Contents:   NAND Flash definitions
//
// $Archive: /USB/atapifx2/NX2LP/SRC/CY3686FW/inand.h $
// $Date: 10/12/05 10:56a $
// $Revision: 3 $
//-----------------------------------------------------------------------------
//
// Copyright 2005, Cypress Semiconductor Corporation.
//
// This software is owned by Cypress Semiconductor Corporation (Cypress)
// and is protected by and subject to worldwide patent protection (United
// States and foreign), United States copyright laws and international 
// treaty provisions. Cypress hereby grants to licensee a personal, 
// non-exclusive, non-transferable license to copy, use, modify, create 
// derivative works of, and compile the Cypress Source Code and derivative 
// works for the sole purpose of creating custom software in support of 
// licensee product to be used only in conjunction with a Cypress integrated 
// circuit as specified in the applicable agreement. Any reproduction, 
// modification, translation, compilation, or representation of this 
// software except as specified above is prohibited without the express 
// written permission of Cypress.
//
// Disclaimer: CYPRESS MAKES NO WARRANTY OF ANY KIND,EXPRESS OR IMPLIED, 
// WITH REGARD TO THIS MATERIAL, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
// Cypress reserves the right to make changes without further notice to the
// materials described herein. Cypress does not assume any liability arising
// out of the application or use of any product or circuit described herein.
// Cypress does not authorize its products for use as critical components in
// life-support systems where a malfunction or failure may reasonably be
// expected to result in significant injury to the user. The inclusion of
// Cypress' product in a life-support systems application implies that the
// manufacturer assumes all risk of such use and in doing so indemnifies
// Cypress against all charges.
//
// Use may be limited by and subject to the applicable Cypress software
// license agreement.
//
//--------------------------------------------------------------------------
// Definition
//
//   USE_2NAND       When enable, it will support only 2 NAND chips
//
//   NAND_2K         enable this define when 2K NAND type is used
//                   disable this define when 512 NAND type is used
//   
//   NO_WP           When enable, it will only support removable device
//  
//   NX2LP           When enable, it will use NX2LP silicon, else it will use
//                   FX2LP silicon
//
//   DEBUG           When enable, will dump data for debugging
//
//   USE_2LUN        Support 2 LUNs
//                   CDROM Type will configured by the Bit1 of the FWCFG
//--------------------------------------------------------------------------

#define cCDROMType   5  // enable CDROM type

#define cLUN1_Capacity  2048  // Allocate 2K sector for LUN1 => 1MB Size

#ifndef INAND_H
#define INAND_H

#define cFail 1
#define cOK   0

// defines for NAND project
sbit NAND_CLE        = 0x80+0;     // PA0
sbit NAND_ALE        = 0x80+1;     // PA1
sbit NAND_LED0       = 0x80+2;     // PA2
sbit NAND_LED1       = 0x80+3;     // PA3
sbit NAND_WP         = 0x80+4;     // PA4
sbit NAND_WP_SWITCH  = 0x80+5;     // PA5


//-----------------------------------------------------------------------------
// Use only 2 NAND chip select
//-----------------------------------------------------------------------------
#ifdef USE_2NAND

sbit CE0             = IOD^0;      // Bank0
sbit CE1             = IOD^1;      // Bank1

#define CE0_ON()     CE0 = 0       // Assert NAND Bank0
#define CE0_OFF()    CE0 = 1       // Negate NAND Bank0

#define CE1_ON()     CE1 = 0       // Assert NAND Bank1
#define CE1_OFF()    CE1 = 1       // Negate NAND Bank1

#define DISABLE_NAND() IOD |= 3    // Negate both Bank0 & Bank1

#else 
//-----------------------------------------------------------------------------
// Use more than 2 NAND chips
//-----------------------------------------------------------------------------

#define DISABLE_NAND() IOD = 0xFF

#endif

#define LED0_ON()    NAND_LED0 = 0
#define LED0_OFF()   NAND_LED0 = 1
#define LED1_ON()    NAND_LED1 = 0
#define LED1_OFF()   NAND_LED1 = 1


//-----------------------------------------------------------------------------
//-----------------------------------------------------------------------------
#define cEP2                 0       // Endpoint FIFO for GPIF
#define cEP4                 1       // Endpoint FIFO for GPIF
#define cEP6                 2       // Endpoint FIFO for GPIF
#define cEP8                 3       // Endpoint FIFO for GPIF

#define EP6CFG_OUT_DEFAULT       0x1a // valid, OUT, ISO, 1024, double buffered
#define EP6CFG_IN_DEFAULT        0x5a // valid, IN, ISO, double buffered, 5a=1024, 52=512 - description corrected
#define EP6FLGSEL_OUT_DEFAULT    0x02 // for EP6 OUTs, use the full flag
#define EP6FLGSEL_IN_DEFAULT     0x01 // for EP6 INs, use the empty flag

// Default GPIF timing will be 60ns
// This macro will enable 40ns
#define GPIF_40()  *(xbyte*)0xe403 = *(xbyte*)0xe423 = 0x01

#ifdef NAND_2K
#define EP4FIFOCFG_DEFAULT      (bmINFM)// | bmAUTOIN)  // Read need to ajdust -1
#else
#define EP4FIFOCFG_DEFAULT		(bmINFM | bmAUTOIN)
#endif

#ifdef NAND_2K
//==========================================================================
// Write Redundant:
//  This subroutine is faster the overhead of Setup the GPIF transfer
//  2K NAND need to finish to write all the un-used byte because the sub-page
//  write
//==========================================================================
#define WriteRedundant()             			\
{                                    			\
   P_XGPIFSGLDATLX = 0xff;			_nop_();	\
   P_XGPIFSGLDATLX = 0xff;          _nop_();	\
   P_XGPIFSGLDATLX = P_ECC1B0[0];	_nop_();	\
   P_XGPIFSGLDATLX = P_ECC1B0[1];	_nop_();	\
   P_XGPIFSGLDATLX = P_ECC1B0[2];	_nop_();	\
   P_XGPIFSGLDATLX = P_ECC1B0[3];	_nop_();	\
   P_XGPIFSGLDATLX = P_ECC1B0[4];	_nop_();	\
   P_XGPIFSGLDATLX = P_ECC1B0[5];	_nop_();	\
   P_XGPIFSGLDATLX = MSB(gDst);		_nop_();	\
   P_XGPIFSGLDATLX = LSB(gDst);		_nop_();	\
   P_XGPIFSGLDATLX = MSB(gDst);		_nop_();	\
   P_XGPIFSGLDATLX = LSB(gDst);		_nop_();	\
   P_XGPIFSGLDATLX = 0xff;			_nop_();	\
   P_XGPIFSGLDATLX = 0xff;			_nop_();	\
   P_XGPIFSGLDATLX = 0xff;			_nop_();	\
   P_XGPIFSGLDATLX = 0xff;			_nop_();	\
}

#else

//==========================================================================
// Write Redundant:
//  This subroutine is faster the overhead of Setup the GPIF transfer
//==========================================================================
#define WriteRedundant()             \
{                                    \
   P_XGPIFSGLDATLX = 0xff;             \
   _nop_();                          \
   P_XGPIFSGLDATLX = 0xff;             \
   P_XGPIFSGLDATLX = P_ECC1B0[0];      \
   P_XGPIFSGLDATLX = P_ECC1B0[1];      \
   P_XGPIFSGLDATLX = P_ECC1B0[2];      \
   P_XGPIFSGLDATLX = P_ECC1B0[3];      \
   P_XGPIFSGLDATLX = P_ECC1B0[4];      \
   P_XGPIFSGLDATLX = P_ECC1B0[5];      \
   P_XGPIFSGLDATLX = MSB(gDst);        \
   P_XGPIFSGLDATLX = LSB(gDst);        \
   P_XGPIFSGLDATLX = MSB(gDst);        \
   P_XGPIFSGLDATLX = LSB(gDst);        \
}
#endif


// the following are the NAND commands for 2K/512 page size NAND
#define cNAND_READ_DATA             0x00
#define cNAND_READ_START            0x30
#define cNAND_READ_REDUNDANT        0x50
#define cNAND_READ_STATUS           0x70
#define cNAND_READ_ID               0x90
#define cNAND_RESET                 0xFF
#define cNAND_ERASE                 0x60
#define cNAND_ERASE_FINISH          0xD0
#define cNAND_WRITE_DATA            0x80
#define cNAND_PROGRAM_PAGE          0x10
#define cNAND_PROGRAM_CACHE         0x15

#define cNAND_STATUS_BIT_FAIL       0x01
#define cNAND_STATUS_BIT_BUSY       0x40
#define cNAND_STATUS_BIT_WP         0x80

// bank0 and bank1 poll for ready
#define nand_ready0()              while (!(P_GPIFREADYSTAT & 1))      // wait for bank0
#define nand_ready1()              while (!(P_GPIFREADYSTAT & 2))      // wait for bank 1
#define nand_ready3()              while ( (P_GPIFREADYSTAT & 3) != 3) // wait for both banks
#define cInterLeaveMsk             63                                // 32 pages * 2 - 1


#ifdef NAND_2K
//==========================================================================
// 2KP NAND macros
//==========================================================================
#define n2k_set_wadd()                                       \
{                                                            \
    nand_ready3();                                           \
    NAND_CLE=1,_nop_();										 \
	P_XGPIFSGLDATLX=cNAND_WRITE_DATA, _nop_();				 \
	NAND_CLE=0; _nop_();									 \
    NAND_ALE = 1, _nop_();									 \
	P_XGPIFSGLDATLX = 0;_nop_();                             \
    P_XGPIFSGLDATLX = 0;_nop_();		                     \
    P_XGPIFSGLDATLX = ((BYTE *)&gPhyAdd)[3];_nop_();		 \
    P_XGPIFSGLDATLX = ((BYTE *)&gPhyAdd)[2];_nop_();		 \
    P_XGPIFSGLDATLX = ((BYTE *)&gPhyAdd)[1];_nop_();		 \
    NAND_ALE = 0;_nop_();		                             \
}

#define n2k_radd(a, msk)                                     \
{                                                            \
    nand_ready3();                                           \
    NAND_CLE=1,_nop_();										 \
	P_XGPIFSGLDATLX=cNAND_READ_DATA, _nop_();				 \
	NAND_CLE=0;_nop_();										 \
    NAND_ALE = 1, _nop_();									 \
	P_XGPIFSGLDATLX = nadd0[msk];_nop_();		     	     \
    P_XGPIFSGLDATLX = nadd1[msk];_nop_();		             \
    P_XGPIFSGLDATLX = ((BYTE *)&a)[3];_nop_();		         \
    P_XGPIFSGLDATLX = ((BYTE *)&a)[2];_nop_();		         \
    P_XGPIFSGLDATLX = ((BYTE *)&a)[1];_nop_();		         \
    NAND_ALE = 0;_nop_();		                             \
    NAND_CLE=1, _nop_();									 \
	P_XGPIFSGLDATLX=cNAND_READ_START, _nop_();				 \
	NAND_CLE=0;_nop_();										 \
}

#define n2k_wadd(a, msk)                                     \
{                                                            \
    NAND_CLE=1,_nop_();										 \
	P_XGPIFSGLDATLX=cNAND_WRITE_DATA, _nop_();				 \
	NAND_CLE=0;_nop_();										 \
    NAND_ALE = 1, _nop_();									 \
	P_XGPIFSGLDATLX = nadd0[msk];_nop_();		             \
    P_XGPIFSGLDATLX = nadd1[msk];_nop_();		             \
    P_XGPIFSGLDATLX = ((BYTE *)&a)[3];_nop_();		         \
    P_XGPIFSGLDATLX = ((BYTE *)&a)[2];_nop_();		         \
    P_XGPIFSGLDATLX = ((BYTE *)&a)[1];_nop_();		         \
    NAND_ALE = 0;_nop_();		                             \
}


#define n2k_set_radd()                                       \
{                                                            \
    NAND_CLE=1,_nop_();										 \
	P_XGPIFSGLDATLX=cNAND_READ_DATA, _nop_();				 \
	NAND_CLE=0; _nop_();									 \
    NAND_ALE = 1, _nop_();									 \
	P_XGPIFSGLDATLX = 0;_nop_();		                     \
    P_XGPIFSGLDATLX = 0;_nop_();		                     \
    P_XGPIFSGLDATLX = ((BYTE *)&gPhyAdd)[3];_nop_();		 \
    P_XGPIFSGLDATLX = ((BYTE *)&gPhyAdd)[2];_nop_();		 \
    P_XGPIFSGLDATLX = ((BYTE *)&gPhyAdd)[1];_nop_();		 \
    NAND_ALE = 0;_nop_();		                             \
    NAND_CLE=1, _nop_();									 \
	P_XGPIFSGLDATLX=cNAND_READ_START, _nop_();				 \
	NAND_CLE=0;_nop_();										 \
}
#endif


//==========================================================================
// 512P NAND Set address
//==========================================================================
#define n512_setadd(c, a)                      \
{                                              \
    NAND_CLE=1,P_XGPIFSGLDATLX=c, NAND_CLE=0;  \
    NAND_ALE = 1, P_XGPIFSGLDATLX = 0;         \
    P_XGPIFSGLDATLX = ((BYTE *)&a)[3];         \
    P_XGPIFSGLDATLX = ((BYTE *)&a)[2];         \
    P_XGPIFSGLDATLX = ((BYTE *)&a)[1];         \
    NAND_ALE = 0;                              \
}     

//==========================================================================
// NAND Erase Block command can be used in both 512P and 2KP
//==========================================================================
#ifdef NAND_2K
#define nand_blk_erase(a)                                \
{                                                        \
	nand_ready3();                                       \
    NAND_CLE=1,_nop_();									 \
	P_XGPIFSGLDATLX=cNAND_ERASE,_nop_();				 \
	NAND_CLE=0;_nop_();									 \
    NAND_ALE = 1;                                        \
	_nop_();											 \
    P_XGPIFSGLDATLX = ((BYTE *)&a)[3];                   \
	_nop_();											 \
    P_XGPIFSGLDATLX = ((BYTE *)&a)[2];                   \
	_nop_();											 \
    P_XGPIFSGLDATLX = ((BYTE *)&a)[1];                   \
	_nop_();											 \
    NAND_ALE = 0;                                        \
	_nop_();											 \
    NAND_CLE=1,_nop_();									 \
	P_XGPIFSGLDATLX=cNAND_ERASE_FINISH,_nop_();			 \
	NAND_CLE=0;_nop_();									 \
	_nop_();_nop_();									 \
}				
#else
#define nand_blk_erase(a)                                \
{                                                        \
    nand_ready3();                                       \
    NAND_CLE=1,P_XGPIFSGLDATLX=cNAND_ERASE, NAND_CLE=0;  \
    NAND_ALE = 1;                                        \
    P_XGPIFSGLDATLX = ((BYTE *)&a)[3];                   \
    P_XGPIFSGLDATLX = ((BYTE *)&a)[2];                   \
    P_XGPIFSGLDATLX = ((BYTE *)&a)[1];                   \
    NAND_ALE = 0;                                        \
    NAND_CLE=1,P_XGPIFSGLDATLX=cNAND_ERASE_FINISH, NAND_CLE=0; \
}     
#endif

//==========================================================================
// NAND write 0x00 for bad sector marking on a page
// can be used where the required variables/constants are defined
// Details:- Write 16 times byte 0x00 starting at Bad Block Msk
//==========================================================================
#ifdef NAND_2K
#define write0ForBadBlock()								\
{														\
	NandSetAdd(cNAND_WRITE_DATA,BadBlockMsk);_nop_();	\
	P_XGPIFSGLDATLX = 0x00; _nop_();						\
	P_XGPIFSGLDATLX = 0x00; _nop_();						\
	P_XGPIFSGLDATLX = 0x00; _nop_();						\
	P_XGPIFSGLDATLX = 0x00; _nop_();						\
	P_XGPIFSGLDATLX = 0x00; _nop_();						\
	P_XGPIFSGLDATLX = 0x00; _nop_();						\
	P_XGPIFSGLDATLX = 0x00; _nop_(); 						\
	P_XGPIFSGLDATLX = 0x00; _nop_();						\
	P_XGPIFSGLDATLX = 0x00; _nop_();						\
	P_XGPIFSGLDATLX = 0x00; _nop_();						\
	P_XGPIFSGLDATLX = 0x00; _nop_();						\
	P_XGPIFSGLDATLX = 0x00; _nop_();						\
	P_XGPIFSGLDATLX = 0x00; _nop_();						\
	P_XGPIFSGLDATLX = 0x00; _nop_();						\
	P_XGPIFSGLDATLX = 0x00; _nop_();						\
	P_XGPIFSGLDATLX = 0x00; _nop_();						\
	nand_send_command(cNAND_PROGRAM_PAGE);					\
	_nop_(); _nop_();										\
}			
#endif
//==========================================================================
// Debug the Log2Phy variables
//==========================================================================
#ifndef DEBUG
#define DBUG_DUMP()
#else
#define DBUG_DUMP()                \
{                                  \     
   *(xbyte*)0xe1e0 = gCurZone;     \
   *(xbyte*)0xe1e1 = z;            \
   *(xbyte*)0xe1e2 = zz;           \
   *(xbyte*)0xe1e3 = page;         \
   *(xword*)0xe1e4 = gDst;         \
   *(xword*)0xe1e6 = gSrc;         \
   *(xdword*)0xe1ec = gPhyAdd;     \
   *(xdword*)0xe1f0 = gSrcAdd;     \
}
#endif

//==========================================================================
// Use FIFO6 as buffer
//==========================================================================
#define fifo6_in()    P_FIFORESET=6, P_EP6CFG=EP6CFG_IN_DEFAULT
#define fifo6_out()   P_FIFORESET=6, P_EP6BCL = P_EP6CFG=EP6CFG_OUT_DEFAULT
#define ARM_EP2()     { P_OUTPKTEND = 2; _nop_(); P_OUTPKTEND = 2; } 


//==========================================================================
// local define for speed optimize
//==========================================================================
#define xLBA3                      (((BYTE *)&dwLBA)[3])
#define xFerLen2                   (((BYTE *)&dataTransferLen)[2])
#define xFerLen1                   (((BYTE *)&dataTransferLen)[1])
#define xSector                    (((BYTE *)&gSectorcount)[1])
#define xPhyAdd                    (((BYTE *)&gPhyAdd)[3])
#define xSrcAdd                    (((BYTE *)&gSrcAdd)[3])
#ifdef NAND_2K
#define nand_send_command(cmd)     {NAND_CLE=1, _nop_();P_XGPIFSGLDATLX=cmd, _nop_(); NAND_CLE=0, _nop_();_nop_();}				
#define nand_get(stat)             {nand_ready3(); NAND_CLE=1, _nop_(); P_XGPIFSGLDATLX=cNAND_READ_STATUS, _nop_(); NAND_CLE=0, _nop_(); nand_ready3(); P_XGPIFSGLDATLX, _nop_(); stat= P_XGPIFSGLDATLNOX;}
#else
#define nand_send_command(cmd)     NAND_CLE=1, P_XGPIFSGLDATLX=cmd, NAND_CLE=0
#define nand_get(stat)             (NAND_CLE=1, P_XGPIFSGLDATLX=cNAND_READ_STATUS, NAND_CLE=0, P_XGPIFSGLDATLX, stat= P_XGPIFSGLDATLNOX) 
#endif

#define writePIO16(e, c)           { P_GPIFTCB1=MSB(c),P_GPIFTCB0= LSB(c), GPIFTRIG=e, _nop_(); while (!gpifIdle()); }
#define readPIO16(e,c)             { P_GPIFTCB1=MSB(c),P_GPIFTCB0= LSB(c), GPIFTRIG=(4|e), _nop_(); while (!gpifIdle()); }
#define FifoRd                     readPIO16
#define FifoWr                     writePIO16

//==========================================================================
// NAND Definition stuff
//==========================================================================
#define cNAND_DSIZE              (512-1)   // adjust for both read/write
#define cNAND_REDUNDANT_SIZE     (16-1)
#define cNAND_PSIZE              (528-1)
#define cNAND_RSIZE               12-1
#ifdef NAND_4K							   // Size of sector for 4k page nand is 1024 bytes and that for 2K/512 is 512 bytes
#define NAND_ATA_SECTOR_SIZE	 1024
#else
#define NAND_ATA_SECTOR_SIZE     512
#endif
#define cCFG_ECC_OFF             0x0a       // ECC offset in the redundant area

#define NAND_SCSI_CMD_VENDOR_CBW 0xC8
#define EP4CFG_DEFAULT           0xE0  // valid, IN, BULK, 512, double
#define EP2CFG_DEFAULT           0xA2  // valid, OUT, BULK, 512, double

                         // The page size has doubled with 4k size, but at the same time the ATA SECTOR Size too and hence the total sectors remain same
#define TOTAL_SECTORS        0x3E800   // 128MB

#define cMaxBlkChk               24     // max 24 blocks
#define cINQUIRY_OFFSET          0x10
#define cINQUIRY_LENGTH          0x20   // INQUIRY Length
#define cVID_OFFSET              0
#define cHALF_BUFF_OFFSET        0x100  // offset for half buffer
#define cSTRING_OFFSET           0x30   // String offset from Vendor Config page
#define cNAND_CFG_LENGTH         0xE0   // From Inquiry offset

#define cMaxBlock      1024       // number of block in a Zone
#define cMaxBlock2N    10         // 2^10 = 1024
#define cMaxLogical    1000       // maximum logical per block
#define cBLK_INIT      0x0400     // block init
#define cBLK_aMSK      (0x03ff|cBLK_INIT) // bit0-10 block address mask: valid 0-1023

#ifdef NAND_2K
#define cMaxZone       8          // maximun zone for 2K NAND is 4 - Changing to 8 based on available products
#define NAND_PCPY      nNandMove
#else
#define cMaxZone       8          // maximun zone for 512 is 8 
#define NAND_PCPY      nCopyPages
#endif

#define cRedundantSize 16
#define cBLK_USE       0x8000     // bit15=1=use else free
#define cBLK_CFG       0x4000     // Config Block or other cases
#define cBLK_ECC       0x2000     // check for repeat ECC error

#define cBLK_BAD       0x4000     // Bad Block
#define cBLK_uMSK      (cBLK_USE|cBLK_CFG)   // both bits are cleared = Block Free

#define c512PageSize   32
#define c2KPageSize    64
#define c2KPageSize2N  6
#define c512PageSize2N 5
#define cAddOffset     8
#define cECCOffset     2
#define c1Gbit_2N      0xd     // 2^13 = 8192
#define c512Mbit_2N    0xc     // 2^12 = 4096

#define ECCBUF         0xE62A
#define EP4FIFO        (0xF400+cECCOffset)
#define cEP6FIFO       (EP6FIFOBUF+512+cECCOffset)

#define cTOSHIBA_ID    0x98
#define cST_ID         0x20
#define cHYNIX_ID      0xad
#define cSAMSUNG_ID    0xec


//==========================================================================
// Local defines from the mass storage class spec
//==========================================================================
#define SC_MASS_STORAGE_RESET       0xff
#define SC_GET_MAX_LUN              0xfe
#define SC_HID_SET_IDLE             0xa

#define CBW_TAG                     4
#define CBW_DATA_TRANSFER_LEN_LSB   8
#define CBW_DATA_TRANSFER_LEN_MSB   9
#define CBW_FLAGS                   12
#define CBW_FLAGS_DIR_BIT           0x80
#define CBW_LUN                     13
#define CBW_CBW_LEN                 14
#define CBW_CBW_LEN_MASK            0xf
#define CBW_DATA_START              15

#define   min(a,b) (((a)<(b))?(a):(b))
#define   max(a,b) (((a)>(b))?(a):(b))

#define senseCRCError               0
#define senseInvalidFieldInCDB      1
#define senseOk                     2
#define senseNoMedia                3
#define senseWriteFault             4
#define senseReadError              5
#define senseAddrNotFound           6
#define senseInvalidOpcode          7
#define senseInvalidLBA             8
#define senseInvalidParameter       9
#define senseCantEject              0xa
#define senseMediaChanged           0xb
#define senseDeviceReset            0xc
#define senseWriteProtected         0xd

#define USBS_PASSED      0
#define USBS_FAILED      1
#define USBS_PHASE_ERROR 2

// Fields in the INQUIRY
#define SCSI_INQUIRY_DEVICE_CLASS      0
#define SCSI_INQUIRY_REMOVABLE_BIT     0x80
#define SCSI_INQUIRY_REMOVABLE_BYTE    1
#define ATAPI_INQUIRY_REMOVABLE_BYTE   0
#define SCSI_INQUIRY_DATA_FORMAT       3
#define SCSI_INQUIRY_MANUFACTURER      8
#define ATAPI_INQUIRY_MANUFACTURER     27
#define SCSI_INQUIRY_MANUFACTURER_LEN  24
#define ATAPI_INQUIRY_REVISION         73
#define SCSI_INQUIRY_REVISION          32
#define SCSI_INQUIRY_REVISION_LEN      4


// SHORT_PACKET_BEFORE_STALL - Determines if a short packet is sent prior to
// the STALL ofan IN endpoint. The USB Mass Storage Class Bulk-Only
// Specification allows a device to send a short or zero-length IN packet
// prior to returning a STALL handshake for certain cases. Certain host
// controller drivers may require a short packet prior to STALL.
#define SHORT_PACKET_BEFORE_STALL   TRUE

#define dataTransferLenLSW	((WORD *) (&dataTransferLen))[1]
#define pStr3Offset ((WORD)(halfKBuffer + (BYTE)&Str3Offset))
#define pVendorOffset ((WORD)(halfKBuffer + (BYTE)&VendorOffset))
#define pHighSpeedConfigDscr ((WORD)(halfKBuffer + (BYTE)&HighSpeedConfigDscrOffset))
#define pDscrVID ((WORD)(halfKBuffer + (BYTE)&DscrVIDOffset))

#define pLSB(offset)   LSB((WORD)(halfKBuffer + (BYTE)&offset))  // general macro usage

#define HS_BULK_PACKET_SIZE     0x200
#define FS_BULK_PACKET_SIZE     0x40
#define BUFFER_SIZE 240         // halfKBuffer buffer size


#endif
